"""Taper definitions [dbu].

TODO: Non-linear tapers
"""

from collections.abc import Callable
from typing import Any, Protocol

from .. import kdb, kf_types
from ..conf import logger
from ..enclosure import LayerEnclosure
from ..kcell import Info, KCell, KCLayout, MetaData, kcl

__all__ = ["taper"]


class TaperFactory(Protocol):
    def __call__(
        self,
        width1: kf_types.dbu,
        width2: kf_types.dbu,
        length: kf_types.dbu,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
    ) -> KCell:
        r"""Linear Taper [dbu].

                   __
                 _/  │ Slab/Exclude
               _/  __│
             _/  _/  │
            │  _/    │
            │_/      │
            │_       │ Core
            │ \_     │
            │_  \_   │
              \_  \__│
                \_   │
                  \__│ Slab/Exclude

        Args:
            width1: Width of the core on the left side. [dbu]
            width2: Width of the core on the right side. [dbu]
            length: Length of the taper. [dbu]
            layer: Main layer of the taper.
            enclosure: Definition of the slab/exclude.
        """
        ...


def taper_factory(
    kcl: KCLayout,
    basename: str | None = None,
    additional_info: Callable[
        ...,
        dict[str, MetaData],
    ]
    | dict[str, MetaData]
    | None = None,
    **cell_kwargs: Any,
) -> TaperFactory:
    r"""Returns a function generating linear tapers [dbu].

               __
             _/  │ Slab/Exclude
           _/  __│
         _/  _/  │
        │  _/    │
        │_/      │
        │_       │ Core
        │ \_     │
        │_  \_   │
          \_  \__│
            \_   │
              \__│ Slab/Exclude

    Args:
        kcl: The KCLayout which will be owned
        additional_info: Add additional key/values to the
            [`KCell.info`][kfactory.kcell.KCell.info]. Can be a static dict
            mapping info name to info value. Or can a callable which takes the straight
            functions' parameters as kwargs and returns a dict with the mapping.
        basename: Overwrite the prefix of the resulting KCell's name. By default
            the KCell will be named 'straight_dbu[...]'.
        cell_kwargs: Additional arguments passed as `@kcl.cell(**cell_kwargs)`.
    """
    if callable(additional_info) and additional_info is not None:
        _additional_info_func: Callable[
            ...,
            dict[str, MetaData],
        ] = additional_info
        _additional_info: dict[str, MetaData] = {}
    else:

        def additional_info_func(
            **kwargs: Any,
        ) -> dict[str, MetaData]:
            return {}

        _additional_info_func = additional_info_func
        _additional_info = additional_info or {}

    @kcl.cell(basename=basename, **cell_kwargs)
    def taper(
        width1: kf_types.dbu,
        width2: kf_types.dbu,
        length: kf_types.dbu,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
    ) -> KCell:
        r"""Linear Taper [um].

                   __
                 _/  │ Slab/Exclude
               _/  __│
             _/  _/  │
            │  _/    │
            │_/      │
            │_       │ Core
            │ \_     │
            │_  \_   │
              \_  \__│
                \_   │
                  \__│ Slab/Exclude

        Args:
            width1: Width of the core on the left side. [dbu]
            width2: Width of the core on the right side. [dbu]
            length: Length of the taper. [dbu]
            layer: Main layer of the taper.
            enclosure: Definition of the slab/exclude.
        """
        c = kcl.kcell()
        if length < 0:
            logger.critical(
                f"Negative lengths are not allowed {length} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            length = -length
        if width1 < 0:
            logger.critical(
                f"Negative widths are not allowed {width1} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            width1 = -width1

        if width2 < 0:
            logger.critical(
                f"Negative widths are not allowed {width2} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            width2 = -width2

        li = c.kcl.find_layer(layer)
        taper = c.shapes(li).insert(
            kdb.Polygon(
                [
                    kdb.Point(0, int(-width1 / 2)),
                    kdb.Point(0, width1 // 2),
                    kdb.Point(length, width2 // 2),
                    kdb.Point(length, int(-width2 / 2)),
                ]
            )
        )

        c.create_port(trans=kdb.Trans(2, False, 0, 0), width=width1, layer=li)
        c.create_port(trans=kdb.Trans(0, False, length, 0), width=width2, layer=li)

        if enclosure is not None:
            enclosure.apply_minkowski_y(c, layer)
        _info: dict[str, MetaData] = {
            "width1_um": width1 * c.kcl.dbu,
            "width2_um": width2 * c.kcl.dbu,
            "length_um": length * c.kcl.dbu,
            "width1_dbu": width1,
            "width2_dbu": width2,
            "length_dbu": length,
        }
        _info.update(
            _additional_info_func(
                width1=width1,
                width2=width2,
                length=length,
                layer=layer,
                enclosure=enclosure,
            )
        )
        _info.update(_additional_info)
        c.info = Info(**_info)
        c.auto_rename_ports()
        c.boundary = taper.dpolygon

        return c

    return taper


taper = taper_factory(kcl)



# ---
# jupyter:
#   jupytext:
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Multi - KCLayout / PDK

# %% [markdown]
# You can also use multiple KCLayout objects as PDKs or Libraries of KCells and parametric KCell-Functions

# %% [markdown]
# ## Use multiple KCLayout objects as PDKs/Libraries
#
# KCLayouts can act as PDKs. They can be seamlessly instantiated into each other

# %%
import kfactory as kf

class LayerInfos(kf.LayerInfos):
    WG: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1,0)
    WGEX: kf.kdb.LayerInfo = kf.kdb.LayerInfo(2,0) # WG Exclude
    CLAD: kf.kdb.LayerInfo = kf.kdb.LayerInfo(4,0) # cladding
    FLOORPLAN: kf.kdb.LayerInfo = kf.kdb.LayerInfo(10,0)

# Make the layout object aware of the new layers:
LAYER = LayerInfos()
kf.kcl.infos = LAYER

kcl_default = kf.kcl

# %% [markdown]
# Empty default KCLayout

# %%
kcl_default.kcells

# %%
# Create a default straight waveguide in the default KCLayout with dbu==0.001 (1nm grid)
s_default = kf.cells.straight.straight(
    width=1, length=10, layer=LAYER.WG
)

# %%
# There is now a a KCell in the KCLayout
kcl_default.kcells

# %%
# Control the dbu is still 1nm
kcl_default.dbu

# %%
# Create a new KCLayout to simulate pdk (could be package with e.g. `from test_pdk import kcl as pdk` or similar)
kcl2 = kf.KCLayout("TEST_PDK", infos=LayerInfos)
# Set the dbu to 0.005 (5nm)
kcl2.dbu = 0.005
kcl2.layout.dbu

# %%
# Since it's a new KCLayout, it's empty
kcl2

# %%
# Create an parametric KCell-Function for straights on the new pdk
sf2 = kf.factories.straight.straight_dbu_factory(kcl=kcl2)

# %%
# Make an instance with
s2 = kcl2.factories["straight"](length=10000, width=200, layer=LAYER.WG)
s2.settings

# %%
# The default kcl's straight uses 1nm grid and is therefore 1000dbu (1um) high and 10000dbu (10um) wide
print(f"{s_default.bbox().height()=}")
print(f"{s_default.dbbox().height()=}")
print(f"{s_default.bbox().width()=}")
print(f"{s_default.dbbox().width()=}")
# The test pdk uses a 5nm grid, so it will be 200dbu (1um) high and 10000dbu (50um) wide
print(f"{s2.bbox().height()=}")
print(f"{s2.dbbox().height()=}")
print(f"{s2.bbox().width()=}")
print(f"{s2.dbbox().width()=}")

# %%
# The ports of the default kcl also have different width dbu dimensions, but are the same in um
s_default.ports.print()
s2.ports.print()
# But in um they are the same
s_default.ports.print(unit="um")
s2.ports.print(unit="um")

# %%
# Both can be instantiated into the same KCell
c = kcl_default.kcell()
si_d = c << s_default
si_2 = c << s2

# %%
si_2.connect("o1", si_d, "o2")

# %%
c



"""Virtual cells for kfactory.

These are used for cases where all angle placement is necessary,
for example in all-angle routing.
"""

from . import circular, euler, straight

__all__ = ["circular", "euler", "straight"]



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.14.5
#   kernelspec:
#     display_name: kernel_name
#     language: python
#     name: kernel_name
# ---

import kfactory as kf


class LayerInfos(kf.LayerInfos):
    SI: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1, 0)
    SIEXCLUDE: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1, 1)


kf.kcl.infos = LayerInfos()
LAYER = LayerInfos()
si_enc = kf.enclosure.LayerEnclosure([(kf.kcl.infos.SIEXCLUDE, 2000)])



import kfactory as kf
import pytest
from random import randint
from functools import partial
from conftest import Layers

from collections.abc import Callable


@pytest.mark.parametrize(
    "x",
    [
        5000,
        0,
    ],
)
def test_route_straight(
    x: int,
    bend90: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
    optical_port: kf.Port,
) -> None:
    c = kf.KCell()
    p1 = optical_port.copy()
    p2 = optical_port.copy()
    p2.trans = kf.kdb.Trans(2, False, x, 0)
    kf.routing.optical.route(
        c,
        p1,
        p2,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90,
    )


@pytest.mark.parametrize(
    "x,y,angle2",
    [
        (20000, 20000, 2),
        (10000, 10000, 3),
        (randint(10001, 20000), randint(10001, 20000), 3),
        (5000, 10000, 3),  # the mean one where points will collide for radius 10000
        (30000, 5000, 3),
        (500, 500, 3),
        (-500, 30000, 3),
        (500, 30000, 3),
        (-10000, 30000, 3),
        (0, 0, 2),
    ],
)
def test_route_bend90(
    bend90: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
    optical_port: kf.Port,
    x: int,
    y: int,
    angle2: int,
) -> None:
    c = kf.KCell()
    p1 = optical_port.copy()
    p2 = optical_port.copy()
    p2.trans = kf.kdb.Trans(angle2, False, x, y)
    b90r = abs(bend90.ports._ports[0].x - bend90.ports._ports[1].x)
    if abs(x) < b90r or abs(y) < b90r:
        kf.config.logfilter.regex = "route is too small, potential collisions:"
    kf.routing.optical.route(
        c,
        p1,
        p2,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90,
    )

    kf.config.logfilter.regex = None


@pytest.mark.parametrize(
    "x,y,angle2",
    [
        (20000, 20000, 2),
        (10000, 10000, 3),
        (randint(10001, 20000), randint(10001, 20000), 3),
        (5000, 10000, 3),  # the mean one where points will collide for radius 10000
        (30000, 5000, 3),
        (500, 500, 3),
        (-500, 30000, 3),
        (500, 30000, 3),
        (-10000, 30000, 3),
        (0, 0, 2),
        (500000, 50000, 2),
    ],
)
def test_route_bend90_invert(
    bend90: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
    optical_port: kf.Port,
    x: int,
    y: int,
    angle2: int,
) -> None:
    c = kf.KCell()
    p1 = optical_port.copy()
    p2 = optical_port.copy()
    p2.trans = kf.kdb.Trans(angle2, False, x, y)
    b90r = abs(bend90.ports._ports[0].x - bend90.ports._ports[1].x)
    if abs(x) < b90r or abs(y) < b90r:
        kf.config.logfilter.regex = "route is too small, potential collisions:"
    kf.routing.optical.route(
        c,
        p1,
        p2,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90,
        route_kwargs={"invert": True},
    )
    kf.config.logfilter.regex = None


@pytest.mark.parametrize(
    "x,y,angle2",
    [
        (40000, 40000, 2),
        (20000, 20000, 3),
        (10000, 10000, 3),
    ],
)
def test_route_bend90_euler(
    bend90_euler: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
    optical_port: kf.Port,
    x: int,
    y: int,
    angle2: int,
) -> None:
    c = kf.KCell()
    p1 = optical_port.copy()
    p2 = optical_port.copy()
    p2.trans = kf.kdb.Trans(angle2, False, x, y)
    b90r = abs(bend90_euler.ports._ports[0].x - bend90_euler.ports._ports[1].x)
    if abs(x) < b90r or abs(y) < b90r:
        kf.config.logfilter.regex = "route is too small, potential collisions:"
    kf.routing.optical.route(
        c,
        p1,
        p2,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_euler,
    )
    kf.config.logfilter.regex = None


def test_route_bundle(
    LAYER: Layers,
    optical_port: kf.Port,
    bend90_euler: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
) -> None:
    c = kf.KCell()

    p_start = [
        optical_port.copy(
            kf.kdb.Trans(
                1,
                False,
                i * 200_000 - 50_000,
                (4 - i) * 6_000 if i < 5 else (i - 5) * 6_000,
            )
        )
        for i in range(10)
    ]
    p_end = [
        optical_port.copy(
            kf.kdb.Trans(3, False, i * 200_000 + i**2 * 19_000 + 500_000, 300_000)
        )
        for i in range(10)
    ]

    c.shapes(kf.kcl.find_layer(10, 0)).insert(
        kf.kdb.Box(-50_000, 0, 1_750_000, -100_000)
    )
    c.shapes(kf.kcl.find_layer(10, 0)).insert(
        kf.kdb.Box(1_000_000, 500_000, p_end[-1].x, 600_000)
    )

    routes = kf.routing.optical.route_bundle(
        c,
        p_start,
        p_end,
        5_000,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_euler,
        on_collision=None,
    )

    for route in routes:
        c.add_port(route.start_port)
        c.add_port(route.end_port)

    c.auto_rename_ports()


def test_route_length(
    bend90_euler: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
    optical_port: kf.Port,
    taper: kf.KCell,
) -> None:
    x, y, angle2 = (70000, 70000, 2)

    c = kf.KCell()
    p1 = optical_port.copy()
    p2 = optical_port.copy()
    p2.trans = kf.kdb.Trans(angle2, False, x, y)
    b90r = abs(bend90_euler.ports._ports[0].x - bend90_euler.ports._ports[1].x)
    if abs(x) < b90r or abs(y) < b90r:
        kf.config.logfilter.regex = "route is too small, potential collisions:"
    route = kf.routing.optical.route(
        c,
        p1,
        p2,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_euler,
        taper_cell=taper,
    )
    kf.config.logfilter.regex = None

    assert route.length == 65196
    assert route.length_straights == 25196
    assert route.length_backbone == 140000
    assert route.n_bend90 == 2


@pytest.mark.parametrize(
    "indirect,sort_ports,start_bbox,start_angle,m2,m1,z,p1,p2",
    [
        (indirect, sort_ports, start_bbox, start_angle, m2, m1, z, p1, p2)
        for indirect in (True, False)
        for sort_ports in (False, True)
        for start_bbox in (False, True)
        for start_angle in (-2, -1, 0, 1, 2)
        for m2 in (True, False)
        for m1 in (True, False)
        for z in (True, False)
        for p1 in (True, False)
        for p2 in (True, False)
    ],
)
def test_smart_routing(
    bend90_small: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    start_bbox: bool,
    sort_ports: bool,
    indirect: bool,
    start_angle: int,
    m2: bool,
    m1: bool,
    z: bool,
    p1: bool,
    p2: bool,
) -> None:
    """Tests all possible smart routing configs."""
    c = kf.KCell(
        f"test_smart_routing_{sort_ports=}_{start_bbox=}_{start_angle=}"
        f"{m2=}_{m1=}_{z=}_{p1=}_{p2=}"
    )
    c.name = c.name.replace("=", "")

    i = 0

    base_t = kf.kdb.Trans.R0

    _port = partial(c.create_port, width=500, layer=kf.kcl.find_layer(1, 0))

    start_ports: list[kf.Port] = []
    end_ports: list[kf.Port] = []
    start_boxes: list[kf.kdb.Box] = []
    end_boxes: list[kf.kdb.Box] = []
    start_bboxes: list[kf.kdb.Box] = []
    end_bboxes: list[kf.kdb.Box] = []

    angles: list[int] = []
    if m2 and (m1 or z):
        angles.append(-2)
    if m1:
        angles.append(-1)
    if z:
        angles.append(0)
    if p1:
        angles.append(1)
    if p2 and (p1 or z):
        angles.append(2)

    for a in range(4):
        t = base_t * kf.kdb.Trans(a // 2 * 3_000_000, a % 2 * 3_000_000)
        start_box = t * kf.kdb.Box(350_000) if start_bbox else kf.kdb.Box()
        end_box = kf.kdb.Box()
        n = 0
        te = (
            kf.kdb.Trans(2, False, -400_000, 400_000)
            if indirect
            else kf.kdb.Trans(-400_000, 0)
        )
        for i in angles:
            angle = a + start_angle + i
            if i == 2:
                for j in range(5):
                    ps = _port(
                        name=f"start_{a=}_{i=}_{j=}",
                        trans=t
                        * kf.kdb.Trans(angle, False, 0, 0)
                        * kf.kdb.Trans(100_000, (1 - j) * 15_000 - 50_000),
                    )
                    pe = _port(
                        name=f"end_{a=}_{i=}_{j=}",
                        trans=t
                        * kf.kdb.Trans(a, False, 0, 0)
                        * te
                        * kf.kdb.Trans(0, (-n - 4 + j * 2) * 40_000 + 600_000),
                    )
                    start_ports.append(ps)
                    end_ports.append(pe)
                    start_box += ps.trans.disp.to_p()
                    end_box += pe.trans.disp.to_p()
                    n += 1
            elif i == -2:
                for j in range(5):
                    ps = _port(
                        name=f"start_{a=}_{i=}_{j=}",
                        trans=t
                        * kf.kdb.Trans(angle, False, 0, 0)
                        * kf.kdb.Trans(100_000, j * 15_000 + 50_000),
                    )
                    pe = _port(
                        name=f"end_{a=}_{i=}_{j=}",
                        trans=t
                        * kf.kdb.Trans(a, False, 0, 0)
                        * te
                        * kf.kdb.Trans(0, -n * 40_000 + 600_000),
                    )
                    start_ports.append(ps)
                    end_ports.append(pe)
                    start_box += ps.trans.disp.to_p()
                    end_box += pe.trans.disp.to_p()
                    n += 1
            else:
                for j in range(10):
                    ps = _port(
                        name=f"start_{a=}_{i=}_{j=}",
                        trans=t
                        * kf.kdb.Trans(angle, False, 0, 0)
                        * kf.kdb.Trans(100_000, j * 15_000 - 50_000),
                    )
                    pe = _port(
                        name=f"end_{a=}_{i=}_{j=}",
                        trans=t
                        * kf.kdb.Trans(a, False, 0, 0)
                        * te
                        * kf.kdb.Trans(0, -n * 40_000 + 600_000),
                    )
                    start_ports.append(ps)
                    end_ports.append(pe)
                    start_box += ps.trans.disp.to_p()
                    end_box += pe.trans.disp.to_p()
                    n += 1

        start_boxes.append(start_box)
        end_boxes.append(end_box)
    for box in start_boxes + end_boxes:
        c.shapes(kf.kcl.find_layer(10, 0)).insert(box)

    for box in start_bboxes:
        c.shapes(kf.kcl.find_layer(11, 0)).insert(box)
    for box in end_bboxes:
        c.shapes(kf.kcl.find_layer(12, 0)).insert(box)

    match (m1, p1):
        case (True, False):
            rf = partial(
                kf.routing.electrical.route_bundle,
                c,
                start_ports=start_ports,
                end_ports=end_ports,
                separation=4000,
                bboxes=start_boxes + end_boxes + start_bboxes + end_bboxes,
                sort_ports=sort_ports,
                bbox_routing="full",
                on_collision="error",
            )
        case (False, True):
            rf = partial(
                kf.routing.electrical.route_bundle_dual_rails,
                c,
                start_ports=start_ports,
                end_ports=end_ports,
                separation=4000,
                bboxes=start_boxes + end_boxes + start_bboxes + end_bboxes,
                sort_ports=sort_ports,
                bbox_routing="full",
                on_collision="error",
                separation_rails=100,
            )
        case _:
            rf = partial(
                kf.routing.optical.route_bundle,
                c,
                start_ports=start_ports,
                end_ports=end_ports,
                separation=4000,
                bboxes=start_boxes + end_boxes + start_bboxes + end_bboxes,
                sort_ports=sort_ports,
                bbox_routing="full",
                on_collision="error",
                bend90_cell=bend90_small,
                straight_factory=straight_factory_dbu,
            )

    # (indirect, sort_ports, start_bbox, start_angle, m2, m1, z, p1, p2)
    match (indirect, sort_ports, start_bbox, start_angle, m2, m1, z, p1, p2):
        case (
            (True, False, False, -1, True, False, False, True, False)
            | (True, False, False, -1, False, False, False, True, False)
            | (True, False, False, 0, False, False, True, False, False)
            | (True, False, False, 1, False, True, False, False, True)
            | (True, False, False, 1, False, True, False, False, False)
            | (True, True, False, -1, True, False, False, True, False)
            | (True, True, False, -1, False, False, False, True, False)
            | (True, True, False, 0, False, False, True, False, False)
            | (True, True, False, 1, False, True, False, False, True)
            | (True, True, False, 1, False, True, False, False, False)
        ):
            with pytest.raises(RuntimeError):  # , match="Routing Collision"):i
                rf()
        case _:
            rf()


def test_custom_router(
    LAYER: Layers,
) -> None:
    c = kf.kcl.kcell("CustomRouter")
    bend90 = kf.cells.circular.bend_circular(width=1, radius=10, layer=LAYER.WG)
    b90r = kf.routing.generic.get_radius(bend90.ports)
    sf = partial(kf.cells.straight.straight_dbu, layer=LAYER.WG)

    start_ports = [
        kf.Port(
            name="in{i}",
            width=1000,
            layer_info=LAYER.WG,
            trans=kf.kdb.Trans(1, False, -850_000 + i * 200_000, 0),
            kcl=c.kcl,
        )
        for i in range(10)
    ]
    end_ports = [
        kf.Port(
            name="in{i}",
            width=1000,
            layer_info=LAYER.WG,
            trans=kf.kdb.Trans(3, False, -400_000 + i * 100_000, 200_000),
            kcl=c.kcl,
        )
        for i in range(10)
    ]

    kf.routing.generic.route_bundle(
        c=c,
        start_ports=start_ports,
        end_ports=end_ports,
        end_straights=50_000,
        start_straights=50_000,
        routing_function=kf.routing.manhattan.route_smart,
        routing_kwargs={
            "bend90_radius": b90r,
            "separation": 4000,
        },
        placer_function=kf.routing.optical.place90,
        placer_kwargs={"bend90_cell": bend90, "straight_factory": sf},
        router_post_process_function=kf.routing.manhattan.path_length_match_manhattan_route,
        router_post_process_kwargs={
            "bend90_radius": b90r,
            "separation": 5000,
        },
    )


def test_route_smart_waypoints_trans_sort(
    bend90_small: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
) -> None:
    c = kf.KCell("TEST_SMART_ROUTE_WAYPOINTS_TRANS_SORT")
    _l = 15
    transformations = [kf.kdb.Trans(0, False, 0, i * 50_000) for i in range(_l)] + [
        kf.kdb.Trans(1, False, -15_000 - i * 50_000, 15 * 50_000) for i in range(_l)
    ]
    start_ports = [
        kf.Port(width=500, layer_info=LAYER.WG, kcl=c.kcl, trans=trans)
        for trans in transformations
    ]
    end_ports = [
        kf.Port(
            width=500,
            layer_info=LAYER.WG,
            kcl=c.kcl,
            trans=kf.kdb.Trans(2, False, 500_000, 0) * trans,
        )
        for trans in transformations
    ]
    kf.routing.optical.route_bundle(
        c,
        start_ports,
        end_ports,
        separation=4000,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_small,
        waypoints=kf.kdb.Trans(250_000, 0),
        sort_ports=True,
    )


def test_route_smart_waypoints_pts_sort(
    bend90_small: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
) -> None:
    c = kf.KCell("TEST_SMART_ROUTE_WAYPOINTS_PTS_SORT")
    _l = 15
    transformations = [kf.kdb.Trans(0, False, 0, i * 50_000) for i in range(_l)] + [
        kf.kdb.Trans(1, False, -15_000 - i * 50_000, 15 * 50_000) for i in range(_l)
    ]
    start_ports = [
        kf.Port(width=500, layer_info=LAYER.WG, kcl=c.kcl, trans=trans)
        for trans in transformations
    ]
    end_ports = [
        kf.Port(
            width=500,
            layer_info=LAYER.WG,
            kcl=c.kcl,
            trans=kf.kdb.Trans(2, False, 500_000, 0) * trans,
        )
        for trans in transformations
    ]
    kf.routing.optical.route_bundle(
        c,
        start_ports,
        end_ports,
        separation=4000,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_small,
        waypoints=[kf.kdb.Point(250_000, 0), kf.kdb.Point(250_000, 100_000)],
        sort_ports=True,
    )


def test_route_smart_waypoints_trans(
    bend90_small: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
) -> None:
    c = kf.KCell("TEST_SMART_ROUTE_WAYPOINTS_TRANS")
    _l = 15
    transformations = [kf.kdb.Trans(0, False, 0, i * 50_000) for i in range(_l)] + [
        kf.kdb.Trans(1, False, -15_000 - i * 50_000, 15 * 50_000) for i in range(_l)
    ]
    start_ports = [
        kf.Port(width=500, layer_info=LAYER.WG, kcl=c.kcl, trans=trans)
        for trans in transformations
    ]
    end_ports = [
        kf.Port(
            width=500,
            layer_info=LAYER.WG,
            kcl=c.kcl,
            trans=kf.kdb.Trans(2, False, 500_000, 0) * trans,
        )
        for trans in transformations
    ]
    kf.routing.optical.route_bundle(
        c,
        start_ports,
        end_ports,
        separation=4000,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_small,
        waypoints=kf.kdb.Trans(250_000, 0),
    )


def test_route_smart_waypoints_pts(
    bend90_small: kf.KCell,
    straight_factory_dbu: Callable[..., kf.KCell],
    LAYER: Layers,
) -> None:
    c = kf.KCell("TEST_SMART_ROUTE_WAYPOINTS_PTS")
    _l = 15
    transformations = [kf.kdb.Trans(0, False, 0, i * 50_000) for i in range(_l)] + [
        kf.kdb.Trans(1, False, -15_000 - i * 50_000, 15 * 50_000) for i in range(_l)
    ]
    start_ports = [
        kf.Port(width=500, layer_info=LAYER.WG, kcl=c.kcl, trans=trans)
        for trans in transformations
    ]
    end_ports = [
        kf.Port(
            width=500,
            layer_info=LAYER.WG,
            kcl=c.kcl,
            trans=kf.kdb.Trans(2, False, 500_000, 0) * trans,
        )
        for trans in transformations
    ]
    kf.routing.optical.route_bundle(
        c,
        start_ports,
        end_ports,
        separation=4000,
        straight_factory=straight_factory_dbu,
        bend90_cell=bend90_small,
        waypoints=[kf.kdb.Point(250_000, 0), kf.kdb.Point(250_000, 100_000)],
    )



"""Virtual circular cell factories."""

from collections.abc import Callable
from typing import Any, Protocol

import numpy as np

from ... import kdb
from ...conf import logger
from ...enclosure import LayerEnclosure
from ...kcell import Info, KCLayout, MetaData, VKCell
from .utils import extrude_backbone

__all__ = ["virtual_bend_circular_factory"]


class BendCircularVKCell(Protocol):
    """Factory for virtual circular bend."""

    def __call__(
        self,
        width: float,
        radius: float,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
        angle: float = 90,
        angle_step: float = 1,
    ) -> VKCell:
        """Create a virtual circular bend.

        Args:
            width: Width of the core. [um]
            radius: Radius of the backbone. [um]
            layer: Layer index of the target layer.
            enclosure: Optional enclosure.
            angle: Angle amount of the bend.
            angle_step: Angle amount per backbone point of the bend.
        """
        ...


def virtual_bend_circular_factory(
    kcl: KCLayout,
    additional_info: Callable[
        ...,
        dict[str, MetaData],
    ]
    | dict[str, MetaData]
    | None = None,
    basename: str | None = None,
    **cell_kwargs: Any,
) -> BendCircularVKCell:
    """Returns a function generating virtual circular bends.

    Args:
        kcl: The KCLayout which will be owned
        additional_info: Add additional key/values to the
            [`VKCell.info`][kfactory.kcell.VKCell.info]. Can be a static dict
            mapping info name to info value. Or can a callable which takes the straight
            functions' parameters as kwargs and returns a dict with the mapping.
        basename: Overwrite the prefix of the resulting VKCell's name. By default
            the VKCell will be named 'virtual_bend_circular[...]'.
        cell_kwargs: Additional arguments passed as `@kcl.vcell(**cell_kwargs)`.
    """
    if callable(additional_info) and additional_info is not None:
        _additional_info_func: Callable[
            ...,
            dict[str, MetaData],
        ] = additional_info
        _additional_info: dict[str, MetaData] = {}
    else:

        def additional_info_func(
            **kwargs: Any,
        ) -> dict[str, MetaData]:
            return {}

        _additional_info_func = additional_info_func
        _additional_info = additional_info or {}

    @kcl.vcell(
        basename=basename,
        **cell_kwargs,
    )
    def virtual_bend_circular(
        width: float,
        radius: float,
        layer: kdb.LayerInfo,
        enclosure: LayerEnclosure | None = None,
        angle: float = 90,
        angle_step: float = 1,
    ) -> VKCell:
        """Create a virtual circular bend.

        Args:
            width: Width of the core. [um]
            radius: Radius of the backbone. [um]
            layer: Layer index of the target layer.
            enclosure: Optional enclosure.
            angle: Angle amount of the bend.
            angle_step: Angle amount per backbone point of the bend.
        """
        c = VKCell()
        if angle < 0:
            logger.critical(
                f"Negative lengths are not allowed {angle} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            angle = -angle
        if width < 0:
            logger.critical(
                f"Negative widths are not allowed {width} as ports"
                " will be inverted. Please use a positive number. Forcing positive"
                " lengths."
            )
            width = -width
        dbu = c.kcl.dbu
        backbone = [
            kdb.DPoint(x, y)
            for x, y in [
                [
                    np.sin(_angle / 180 * np.pi) * radius,
                    (-np.cos(_angle / 180 * np.pi) + 1) * radius,
                ]
                for _angle in np.linspace(
                    0, angle, int(angle // angle_step + 0.5), endpoint=True
                )
            ]
        ]

        extrude_backbone(
            c=c,
            backbone=backbone,
            width=width,
            layer=layer,
            enclosure=enclosure,
            start_angle=0,
            end_angle=angle,
            dbu=dbu,
        )
        _info: dict[str, MetaData] = {}
        _info.update(
            _additional_info_func(
                width=width,
                radius=radius,
                layer=layer,
                enclosure=enclosure,
                angle=angle,
                angle_step=angle_step,
            )
        )
        _info.update(_additional_info)
        c.info = Info(**_info)

        c.create_port(
            name="o1",
            layer=c.kcl.find_layer(layer),
            dwidth=round(width / c.kcl.dbu),
            dcplx_trans=kdb.DCplxTrans(1, 180, False, backbone[0].to_v()),
        )
        c.create_port(
            name="o2",
            dcplx_trans=kdb.DCplxTrans(1, angle, False, backbone[-1].to_v()),
            dwidth=width,
            layer=c.kcl.find_layer(layer),
        )
        return c

    return virtual_bend_circular



import kfactory as kf

import pytest

from functools import partial
from tempfile import NamedTemporaryFile
from pathlib import Path
from conftest import Layers


def test_pdk(LAYER: Layers) -> None:
    pdk = kf.KCLayout("PDK")

    # class LAYER(kf.kcell.LayerEnum):
    #     kcl = kf.constant(pdk)
    #     WG = (1, 0)
    #     WGEX = (1, 1)
    class LAYERS(kf.LayerInfos):
        WG: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1, 0)
        WGEX: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1, 1)

    pdk.infos = LAYER
    for layer in LAYER.model_dump().values():
        assert getattr(pdk.layers, layer.name).layer == layer.layer


def test_clear(LAYER: Layers) -> None:
    kcl = kf.KCLayout("CLEAR")
    kcl.layer(500, 0)
    kcl.infos = kf.LayerInfos(**{"WG": kf.kdb.LayerInfo(1, 0)})
    # kcl.layers = kcl.layerenum_from_dict(layers=LAYER)
    # kcl.layers = kcl.layerenum_from_dict(
    #     layers=kf.LayerInfos(WG=kf.kdb.LayerInfo(1, 0))
    # )
    assert kcl.layers.WG == 1
    kcl.clear(keep_layers=True)
    assert kcl.layers.WG == 0


def test_kcell_delete(LAYER: Layers) -> None:
    _kcl = kf.KCLayout("DELETE", infos=Layers)

    _kcl.layers = _kcl.layerenum_from_dict(layers=LAYER)

    s = partial(
        kf.factories.straight.straight_dbu_factory(_kcl),
        width=1000,
        length=10_000,
        layer=LAYER.WG,
    )

    s1 = s()
    _kcl.delete_cell(s1)
    assert s1._destroyed() is True

    s1 = s()
    assert s1._destroyed() is False


def test_multi_pdk(LAYER: Layers) -> None:
    base_pdk = kf.KCLayout("BASE", infos=Layers)

    doe_pdk1 = kf.KCLayout(name="DOE1", infos=Layers)
    doe_pdk2 = kf.KCLayout(name="DOE2", infos=Layers)
    assembly_pdk = kf.KCLayout(name="ASSEMBLY", infos=Layers)

    wg = kf.factories.straight.straight_dbu_factory(base_pdk)
    bend90_pdk1 = kf.factories.euler.bend_euler_factory(doe_pdk1)
    bend90_pdk2 = kf.factories.euler.bend_euler_factory(doe_pdk2)

    doe1 = doe_pdk1.kcell("DOE1")
    doe1_wg = doe1 << wg(width=1000, length=10_000, layer=LAYER.WG)
    doe_b1 = doe1 << bend90_pdk1(width=1, radius=10, layer=LAYER.WG)
    doe_b1.connect("o1", doe1_wg, "o2")

    doe1.add_port(doe1_wg.ports["o1"], name="o1")
    doe1.add_port(doe_b1.ports["o2"], name="o2")

    doe2 = doe_pdk2.kcell("DOE2")
    enc2 = kf.LayerEnclosure(
        name="enc2",
        sections=[(LAYER.WGCLAD, 0, 2000)],
        main_layer=LAYER.WG,
    )
    doe2_wg = doe2 << wg(width=1000, length=10_000, layer=LAYER.WG)
    doe_b2 = doe2 << bend90_pdk2(width=1, radius=10, layer=LAYER.WG)
    doe_b2.connect("o1", doe2_wg, "o2")
    doe_b3 = doe2 << bend90_pdk2(width=1, radius=10, layer=LAYER.WG, enclosure=enc2)

    doe_b2.connect("o1", doe2_wg, "o2")
    doe_b3.connect("o1", doe_b2, "o2")

    doe2.add_port(doe2_wg.ports["o1"], name="o1")
    doe2.add_port(doe_b3.ports["o2"], name="o2")

    assembly = assembly_pdk.kcell("TOP")
    d1 = assembly << doe1
    d2 = assembly << doe2
    d2.connect("o1", d1, "o2")


def test_multi_pdk_convert(LAYER: Layers) -> None:
    base_pdk = kf.KCLayout("BASE", infos=Layers)

    doe_pdk1 = kf.KCLayout(name="DOE1", infos=Layers)
    doe_pdk2 = kf.KCLayout(name="DOE2", infos=Layers)
    assembly_pdk = kf.KCLayout(name="ASSEMBLY", infos=Layers)

    wg = kf.factories.straight.straight_dbu_factory(base_pdk)
    bend90_pdk1 = kf.factories.euler.bend_euler_factory(doe_pdk1)
    bend90_pdk2 = kf.factories.euler.bend_euler_factory(doe_pdk2)

    doe1 = doe_pdk1.kcell("DOE1")
    doe1_wg = doe1 << wg(width=1000, length=10_000, layer=LAYER.WG)
    doe_b1 = doe1 << bend90_pdk1(width=1, radius=10, layer=LAYER.WG)
    doe_b1.connect("o1", doe1_wg, "o2")

    doe1.add_port(doe1_wg.ports["o1"], name="o1")
    doe1.add_port(doe_b1.ports["o2"], name="o2")

    doe2 = doe_pdk2.kcell("DOE2")
    enc2 = kf.LayerEnclosure(
        name="enc2",
        sections=[(LAYER.WGCLAD, 0, 2000)],
        main_layer=LAYER.WG,
    )
    doe2_wg = doe2 << wg(width=1000, length=10_000, layer=LAYER.WG)
    doe_b2 = doe2 << bend90_pdk2(width=1, radius=10, layer=LAYER.WG)
    doe_b2.connect("o1", doe2_wg, "o2")
    doe_b3 = doe2 << bend90_pdk2(width=1, radius=10, layer=LAYER.WG, enclosure=enc2)

    doe_b2.connect("o1", doe2_wg, "o2")
    doe_b3.connect("o1", doe_b2, "o2")

    doe2.add_port(doe2_wg.ports["o1"], name="o1")
    doe2.add_port(doe_b3.ports["o2"], name="o2")

    assembly = assembly_pdk.kcell("TOP")
    d1 = assembly << doe1
    d2 = assembly << doe2
    d2.connect("o1", d1, "o2")

    p = Path("ASSEMBLY.oas")
    assembly.write(p, convert_external_cells=True)
    p.unlink()


def test_multi_pdk_read_write(LAYER: Layers) -> None:
    base_pdk = kf.KCLayout("BASE", infos=Layers)

    doe_pdk1_write = kf.KCLayout(name="DOE1_WRITE", infos=Layers)
    doe_pdk2_write = kf.KCLayout(name="DOE2_WRITE", infos=Layers)
    assembly_pdk = kf.KCLayout(name="ASSEMBLY", infos=Layers)

    wg = kf.factories.straight.straight_dbu_factory(base_pdk)
    bend90_pdk1 = kf.factories.euler.bend_euler_factory(doe_pdk1_write)
    bend90_pdk2 = kf.factories.euler.bend_euler_factory(doe_pdk2_write)

    doe1 = doe_pdk1_write.kcell("DOE1")
    doe1_wg = doe1 << wg(width=1000, length=10_000, layer=LAYER.WG)
    doe_b1 = doe1 << bend90_pdk1(width=1, radius=10, layer=LAYER.WG)
    doe_b1.connect("o1", doe1_wg, "o2")

    doe1.add_port(doe1_wg.ports["o1"], name="o1")
    doe1.add_port(doe_b1.ports["o2"], name="o2")

    doe2 = doe_pdk2_write.kcell("DOE2")
    enc2 = kf.LayerEnclosure(
        name="enc2",
        sections=[(LAYER.WGCLAD, 0, 2000)],
        main_layer=LAYER.WG,
    )
    doe2_wg = doe2 << wg(width=1000, length=10_000, layer=LAYER.WG)
    doe_b2 = doe2 << bend90_pdk2(width=1, radius=10, layer=LAYER.WG)
    doe_b2.connect("o1", doe2_wg, "o2")
    doe_b3 = doe2 << bend90_pdk2(width=1, radius=10, layer=LAYER.WG, enclosure=enc2)

    doe_b2.connect("o1", doe2_wg, "o2")
    doe_b3.connect("o1", doe_b2, "o2")

    doe2.add_port(doe2_wg.ports["o1"], name="o1")
    doe2.add_port(doe_b3.ports["o2"], name="o2")

    doe_pdk1_read = kf.KCLayout("DOE1_READ", infos=Layers)
    doe_pdk2_read = kf.KCLayout("DOE2_READ", infos=Layers)

    with NamedTemporaryFile(suffix=".gds") as tf:
        doe_pdk1_write.write(tf.name)
        doe_pdk1_read.read(tf.name)
    with NamedTemporaryFile(suffix=".gds") as tf:
        doe_pdk2_write.write(tf.name)
        doe_pdk2_read.read(tf.name)

    assembly = assembly_pdk.kcell("TOP")
    d1 = assembly << doe_pdk1_read[doe1.name]
    d2 = assembly << doe_pdk2_read[doe2.name]
    d2.connect("o1", d1, "o2")


def test_merge_read_shapes(LAYER: Layers) -> None:
    with pytest.raises(kf.kcell.MergeError):
        kcl_1 = kf.KCLayout("MERGE_BASE", infos=Layers)
        s_base = kf.factories.straight.straight_dbu_factory(kcl_1)(
            width=1000, length=10_000, layer=LAYER.WG
        )
        s_copy = s_base.dup()
        s_copy.name = "Straight"

        kcl_2 = kf.KCLayout("MERGE_READ", infos=Layers)
        kcl_2.layers = kcl_2.layerenum_from_dict(layers=LAYER)
        s_base = kf.factories.straight.straight_dbu_factory(kcl_2)(
            width=1100, length=10_000, layer=LAYER.WG
        )
        s_copy = s_base.dup()
        s_copy.name = "Straight"

        kcl_2.write("MERGE_READ.oas")
        kf.config.logfilter.regex = "(?:Found poly)|(?:MetaInfo 'kfactory:)"
        kcl_1.read("MERGE_READ.oas")
        kf.config.logfilter.regex = None
    Path("MERGE_READ.oas").unlink(missing_ok=True)


def test_merge_read_instances(LAYER: Layers) -> None:
    with pytest.raises(kf.kcell.MergeError):
        kcl_1 = kf.KCLayout("MERGE_BASE", infos=Layers)
        kcl_1.layers = kcl_1.layerenum_from_dict(layers=LAYER)

        enc1 = kf.LayerEnclosure(sections=[(LAYER.WG, 0, 200)], name="CLAD")
        s_base = kf.factories.straight.straight_dbu_factory(kcl_1)(
            width=1000, length=10_000, layer=LAYER.WGEXCLUDE, enclosure=enc1
        )
        s_copy = kcl_1.kcell("Straight")
        s_copy << s_base

        kcl_2 = kf.KCLayout("MERGE_READ", infos=Layers)
        kcl_2.layers = kcl_2.layerenum_from_dict(layers=LAYER)
        enc2 = kf.LayerEnclosure(sections=[(LAYER.WG, 0, 200)], name="CLAD")
        s_base = kf.factories.straight.straight_dbu_factory(kcl_2)(
            width=1000, length=10_000, layer=LAYER.WGEXCLUDE, enclosure=enc2
        )
        s_copy = kcl_2.kcell("Straight")
        copy = s_copy << s_base
        copy.movey(-500)

        kcl_2.write("MERGE_READ.oas")
        kf.config.logfilter.regex = "Found instance"
        kcl_1.read("MERGE_READ.oas")
        kf.config.logfilter.regex = None
    Path("MERGE_READ.oas").unlink(missing_ok=True)


def test_merge_properties() -> None:
    with pytest.raises(kf.kcell.MergeError):
        kcl_1 = kf.KCLayout("MERGE_BASE", infos=Layers)
        c = kcl_1.kcell("properties_cell")
        c.info["test_prop"] = "kcl_1"

        kcl_2 = kf.KCLayout("MERGE_READ", infos=Layers)
        c = kcl_2.kcell("properties_cell")
        c.info["test_prop"] = "kcl_2"

        kcl_2.write("MERGE_READ.oas")
        regex = kf.config.logfilter.regex
        kf.config.logfilter.regex = (
            "MetaInfo differs between existing 'kcl_1' and loaded 'kcl_2'"
        )
        kcl_1.read("MERGE_READ.oas")
        kf.config.logfilter.regex = regex
        Path("MERGE_READ.oas").unlink(missing_ok=True)


def test_pdk_cell_infosettings(straight: kf.KCell, LAYER: Layers) -> None:
    kcl = kf.KCLayout("INFOSETTINGS", infos=Layers)
    c = kcl.kcell()
    _wg = c << straight
    _wg.cell
    assert _wg.cell.settings == straight.settings
    assert _wg.cell.info == straight.info



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.16.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

from layers import LAYER, si_enc
from straight import straight

import kfactory as kf


@kf.cell
def composite_cell() -> kf.KCell:
    c = kf.KCell()

    bend = c.create_inst(
        kf.cells.circular.bend_circular(
            1000 * c.kcl.dbu, 20000 * c.kcl.dbu, LAYER.SI, enclosure=si_enc
        )  # the standard kf.cells are in um, so we need to convert to dbu
    )
    wg = c << straight(1000, 5000, 5000)

    wg.connect("o1", bend, "o2")

    c.add_port(name="1", port=bend.ports["o1"])
    c.add_port(name="2", port=wg.ports["o2"])

    c.auto_rename_ports()

    c.draw_ports()

    return c


if __name__ == "__main__":
    kf.show(composite_cell())



"""Euler bends.

Euler bends are bends with a constantly changing radius
from zero to a maximum radius and back to 0 at the other
end.

There are two kinds of euler bends. One that snaps the ports and one that doesn't.
All the default bends use snapping. To use no snapping make an instance of
BendEulerCustom(KCell.kcl) and use that one.
"""

from ..factories.euler import bend_euler_factory, bend_s_euler_factory
from ..kcell import kcl

__all__ = [
    "bend_euler",
    "bend_s_euler",
]


bend_euler = bend_euler_factory(kcl)
bend_s_euler = bend_s_euler_factory(kcl)



"""Virtual euler cells."""

from ...factories.virtual.euler import virtual_bend_euler_factory
from ...kcell import kcl

__all__ = ["virtual_bend_euler"]

virtual_bend_euler = virtual_bend_euler_factory(kcl)



"""Bezier curve based bends and functions."""

from collections.abc import Callable, Sequence
from typing import Any, Protocol

import numpy as np
import numpy.typing as nty
from scipy.special import binom  # type:ignore[import-untyped,unused-ignore]

from .. import kdb, kf_types
from ..enclosure import LayerEnclosure
from ..kcell import Info, KCell, KCLayout, MetaData

__all__ = ["bend_s_bezier_factory"]


class BezierKCell(Protocol):
    def __call__(
        self,
        width: kf_types.um,
        height: kf_types.um,
        length: kf_types.um,
        layer: kdb.LayerInfo,
        nb_points: int = 99,
        t_start: float = 0,
        t_stop: float = 1,
        enclosure: LayerEnclosure | None = None,
    ) -> KCell:
        """Creat a bezier bend.

        Args:
            width: Width of the core. [um]
            height: height difference of left/right. [um]
            length: Length of the bend. [um]
            layer: Layer index of the core.
            nb_points: Number of points of the backbone.
            t_start: start
            t_stop: end
            enclosure: Slab/Exclude definition. [dbu]
        """
        ...


def bezier_curve(
    t: nty.NDArray[np.float64],
    control_points: Sequence[tuple[np.float64 | float, np.float64 | float]],
) -> list[kdb.DPoint]:
    """Calculates the backbone of a bezier bend."""
    xs = np.zeros(t.shape, dtype=np.float64)
    ys = np.zeros(t.shape, dtype=np.float64)
    n = len(control_points) - 1
    for k in range(n + 1):
        ank = binom(n, k) * (1 - t) ** (n - k) * t**k
        xs += ank * control_points[k][0]
        ys += ank * control_points[k][1]

    return [kdb.DPoint(float(x), float(y)) for x, y in zip(xs, ys)]


def bend_s_bezier_factory(
    kcl: KCLayout,
    additional_info: Callable[
        ...,
        dict[str, MetaData],
    ]
    | dict[str, MetaData]
    | None = None,
    basename: str | None = None,
    **cell_kwargs: Any,
) -> BezierKCell:
    """Returns a function generating bezier s-bends.

    Args:
        kcl: The KCLayout which will be owned
        additional_info: Add additional key/values to the
            [`KCell.info`][kfactory.kcell.KCell.info]. Can be a static dict
            mapping info name to info value. Or can a callable which takes the straight
            functions' parameters as kwargs and returns a dict with the mapping.
        basename: Overwrite the prefix of the resulting KCell's name. By default
            the KCell will be named 'straight_dbu[...]'.
        cell_kwargs: Additional arguments passed as `@kcl.cell(**cell_kwargs)`.
    """
    if callable(additional_info) and additional_info is not None:
        _additional_info_func: Callable[
            ...,
            dict[str, MetaData],
        ] = additional_info
        _additional_info: dict[str, MetaData] = {}
    else:

        def additional_info_func(
            **kwargs: Any,
        ) -> dict[str, MetaData]:
            return {}

        _additional_info_func = additional_info_func
        _additional_info = additional_info or {}

    @kcl.cell(basename=basename, **cell_kwargs)
    def bend_s_bezier(
        width: kf_types.um,
        height: kf_types.um,
        length: kf_types.um,
        layer: kdb.LayerInfo,
        nb_points: int = 99,
        t_start: float = 0,
        t_stop: float = 1,
        enclosure: LayerEnclosure | None = None,
    ) -> KCell:
        """Creat a bezier bend.

        Args:
            width: Width of the core. [um]
            height: height difference of left/right. [um]
            length: Length of the bend. [um]
            layer: Layer index of the core.
            nb_points: Number of points of the backbone.
            t_start: start
            t_stop: end
            enclosure: Slab/Exclude definition. [dbu]
        """
        c = kcl.kcell()
        _length, _height = length, height
        pts = bezier_curve(
            control_points=[
                (0.0, 0.0),
                (_length / 2, 0.0),
                (_length / 2, _height),
                (_length, _height),
            ],
            t=np.linspace(t_start, t_stop, nb_points),
        )

        if enclosure is None:
            enclosure = LayerEnclosure()

        enclosure.extrude_path(
            c, path=pts, main_layer=layer, width=width, start_angle=0, end_angle=0
        )

        c.create_port(
            width=int(width / c.kcl.dbu),
            trans=kdb.Trans(2, False, 0, 0),
            layer=c.kcl.find_layer(layer),
            port_type="optical",
        )
        c.create_port(
            width=int(width / c.kcl.dbu),
            trans=kdb.Trans(
                0, False, c.bbox().right, c.bbox().top - int(width / c.kcl.dbu) // 2
            ),
            layer=c.kcl.find_layer(layer),
            port_type="optical",
        )
        _info: dict[str, MetaData] = {}
        _info.update(
            _additional_info_func(
                width=width,
                height=height,
                length=length,
                layer=layer,
                nb_points=nb_points,
                t_start=t_start,
                t_stop=t_stop,
                enclosure=enclosure,
            )
        )
        _info.update(_additional_info)
        c.info = Info(**_info)

        c.auto_rename_ports()

        return c

    return bend_s_bezier



"""Circular bends.

A circular bend has a constant radius.
"""

from ..factories.circular import bend_circular_factory
from ..kcell import kcl

__all__ = ["bend_circular"]


bend_circular = bend_circular_factory(kcl)
"""Circular bend on the default KCLayout."""



import kfactory as kf
from conftest import Layers


def test_l2n(LAYER: Layers) -> None:
    c = kf.KCell("netlist_test")

    b = kf.cells.circular.bend_circular(width=1, radius=10, layer=LAYER.WG)
    s = kf.cells.straight.straight(width=1, length=20, layer=LAYER.WG)

    b1 = c << b
    b2 = c << b
    b2.connect("o1", b1, "o2")
    b3 = c << b
    b3.connect("o1", b2, "o2")
    s1 = c << s
    s1.connect("o1", b3, "o2")
    c.add_port(b1.ports["o1"])
    p = s1.ports["o2"].copy()
    p.name = None
    c.add_port(p)

    c.l2n()



import kfactory as kf
import pytest
from conftest import Layers


@kf.cell
def mmi_enc(layer: kf.kdb.LayerInfo, enclosure: kf.LayerEnclosure) -> kf.KCell:
    c = kf.KCell()
    li = c.kcl.find_layer(layer)
    c.shapes(li).insert(kf.kdb.Box(-10000, -6000, 10000, 6000))

    taper = kf.kdb.Polygon(
        [
            kf.kdb.Point(0, -500),
            kf.kdb.Point(0, 500),
            kf.kdb.Point(2000, 250),
            kf.kdb.Point(2000, -250),
        ]
    )

    for t in [
        kf.kdb.Trans(0, False, 10000, -4000),
        kf.kdb.Trans(0, False, 10000, 4000),
        kf.kdb.Trans(2, False, -10000, -4000),
        kf.kdb.Trans(2, False, -10000, 4000),
    ]:
        c.shapes(li).insert(taper.transformed(t))

    enclosure.apply_minkowski_enc(c, layer)

    return c


def test_enclosure(LAYER: Layers) -> None:
    kf.LayerEnclosure([(LAYER.WG, 500, -250)])


def test_enc(LAYER: Layers, wg_enc: kf.LayerEnclosure) -> None:
    wg_enc

    mmi_enc(LAYER.WG, wg_enc)


def test_neg_enc(LAYER: Layers) -> None:
    enc = kf.LayerEnclosure([(LAYER.WGCLAD, -1500, 1000)])

    mmi_enc(LAYER.WG, enc)


def test_layer_multi_enc(LAYER: Layers) -> None:
    enc = kf.LayerEnclosure(
        [
            (LAYER.WGCLAD, -5000, -5400),
            (LAYER.WGCLAD, -4000, -3900),
            (LAYER.WGCLAD, -100, 100),
            (LAYER.WGCLAD, -500, -400),
        ]
    )
    mmi_enc(LAYER.WG, enc)


def test_bbox_enc(LAYER: Layers) -> None:
    enc = kf.LayerEnclosure(
        [
            (LAYER.WGCLAD, -5000, -5400),
            (LAYER.WGCLAD, -4000, -3900),
            (LAYER.WGCLAD, -100, 100),
            (LAYER.WGCLAD, -500, -400),
        ],
        main_layer=LAYER.WG,
    )
    c = kf.KCell("BBOX_ENC")
    enc.apply_bbox(c, ref=LAYER.WG)


def test_layer_merge_enc(LAYER: Layers) -> None:
    enc = kf.LayerEnclosure(
        [
            (LAYER.WGCLAD, -5000, -3000),
            (LAYER.WGCLAD, -4000, -2000),
            (LAYER.WGCLAD, -2000, 1000),
        ]
    )
    mmi_enc(LAYER.WG, enc)


def test_um_enclosure(LAYER: Layers) -> None:
    enc = kf.LayerEnclosure(
        [
            (LAYER.WGCLAD, -5000, -3000),
            (LAYER.WGCLAD, -4000, -2000),
            (LAYER.WGCLAD, -2000, 1000),
        ],
        kcl=kf.kcl,
    )

    enc_um = kf.LayerEnclosure(
        dsections=[
            (LAYER.WGCLAD, -5, -3),
            (LAYER.WGCLAD, -4, -2),
            (LAYER.WGCLAD, -2, 1),
        ],
        kcl=kf.kcl,
    )

    assert enc == enc_um


def test_um_enclosure_nodbu(LAYER: Layers) -> None:
    """When defining um sections, kcl must be defined."""
    with pytest.raises(AssertionError):
        kf.LayerEnclosure(
            dsections=[
                (LAYER.WGCLAD, -5, -3),
                (LAYER.WGCLAD, -4, -2),
                (LAYER.WGCLAD, -2, 1),
            ]
        )


def test_pdkenclosure(LAYER: Layers, straight_blank: kf.KCell) -> None:
    c = kf.KCell("wg_slab")

    wg_box = kf.kdb.Box(10000, 500)
    c.shapes(c.kcl.find_layer(LAYER.WG)).insert(wg_box)
    c.shapes(c.kcl.find_layer(LAYER.WGCLAD)).insert(wg_box.enlarged(0, 2500))
    c.create_port(
        trans=kf.kdb.Trans(0, False, wg_box.right, 0),
        width=wg_box.height(),
        layer=c.kcl.find_layer(LAYER.WG),
    )
    c.create_port(
        trans=kf.kdb.Trans(2, False, wg_box.left, 0),
        width=wg_box.height(),
        layer=c.kcl.find_layer(LAYER.WG),
    )

    enc1 = kf.LayerEnclosure(
        sections=[
            (LAYER.WGEXCLUDE, 1000),
        ],
        name="WGEX",
        main_layer=LAYER.WG,
    )

    enc2 = kf.LayerEnclosure(
        name="CLADEX",
        main_layer=LAYER.WGCLAD,
        sections=[(LAYER.WGEXCLUDE, 1000), (LAYER.WGCLADEXCLUDE, 2000)],
    )

    pdkenc = kf.KCellEnclosure(enclosures=[enc1, enc2])

    pdkenc.apply_minkowski_tiled(c, carve_out_ports=True)

    port_wg_ex = kf.kdb.Region()
    box = kf.kdb.Polygon(
        kf.kdb.Box(
            0,
            -wg_box.height() // 2 - 1000,
            wg_box.height() // 2 + 1000,
            wg_box.height() // 2 + 1000,
        )
    )
    for port in c.ports:
        port_wg_ex.insert(box.transformed(port.trans))

    port_wg_ex.merge()

    assert (
        kf.kdb.Region(c.shapes(c.kcl.find_layer(LAYER.WGEXCLUDE))) & port_wg_ex
    ).is_empty()
    assert (
        (kf.kdb.Region(c.shapes(c.kcl.find_layer(LAYER.WGCLADEXCLUDE))) & port_wg_ex)
        - port_wg_ex
    ).is_empty()



"""Tests for read and write of metadata."""

from tempfile import NamedTemporaryFile

import kfactory as kf
from pathlib import Path
from conftest import Layers


@kf.cell  # type: ignore[misc, unused-ignore]
def sample(
    s: str = "a", i: int = 3, f: float = 2.0, t: tuple[int, ...] = (1,)
) -> kf.KCell:
    c = kf.KCell()
    c.info["s"] = s
    c.info["i"] = i
    c.info["f"] = f
    c.info["t"] = t
    c.info["d"] = {
        "a": 1,
        "b": 2,
        "c": {
            "dbox": kf.kdb.DBox(5),
            "polygon": kf.kdb.Polygon(
                pts=[kf.kdb.Point(0, 0), kf.kdb.Point(500, 0), kf.kdb.Point(250, 250)]
            ),
        },
    }
    c.info["e"] = None
    c.write("_test_meta_sample.oas")

    kcl2 = kf.KCLayout("TEST_META_SAMPLE")
    kcl2.read("_test_meta_sample.oas")
    Path("_test_meta_sample.oas").unlink()
    assert kcl2[c.name].info["d"]["c"]["dbox"] == c.info["d"]["c"]["dbox"]
    assert kcl2[c.name].info["d"]["c"]["polygon"] == c.info["d"]["c"]["polygon"]
    assert kcl2[c.name].info["e"] is None
    return c


def test_settings_and_info() -> None:
    c = sample()
    assert c.info["s"] == "a"
    assert c.info["i"] == 3
    assert c.info["f"] == 2.0
    assert c.info["t"] == (1,)
    assert c.settings["s"] == "a"
    assert c.settings["i"] == 3
    assert c.settings["f"] == 2.0
    assert c.settings["t"] == (1,)

    assert "s" in c.info
    assert "s" in c.settings

    c = sample(s="b", i=4, f=3.0, t=(2, 3))
    assert c.info["s"] == "b"
    assert c.info["i"] == 4
    assert c.info["f"] == 3.0
    assert c.info["t"] == (2, 3)
    assert c.settings["s"] == "b"
    assert c.settings["i"] == 4
    assert c.settings["f"] == 3.0
    assert c.settings["t"] == (2, 3)

    c.info["None"] = None


def test_metainfo_set(straight: kf.KCell) -> None:
    """Test autogenerated port metadata."""
    ports = straight.ports.copy()

    straight._locked = False
    straight.set_meta_data()

    straight.ports = kf.Ports(kcl=straight.kcl)

    straight.get_meta_data()

    for i, port in enumerate(ports):
        meta_port = straight.ports[i]

        assert port.name == meta_port.name
        assert port.width == meta_port.width
        assert port.trans == meta_port.trans
        assert port.dcplx_trans == meta_port.dcplx_trans
        assert port.port_type == meta_port.port_type


def test_metainfo_read(LAYER: Layers, straight: kf.KCell) -> None:
    """Test whether we can read written metadata to ports."""
    with NamedTemporaryFile("a", suffix=".oas") as t:
        save = kf.save_layout_options()
        save.write_context_info = True
        straight.kcl.write(t.name)

        kcl = kf.KCLayout("TEST_META", infos=Layers)
        kcl.read(t.name)

        wg_read = kcl[straight.name]
        wg_read.get_meta_data()
        for i, port in enumerate(straight.ports):
            read_port = wg_read.ports[i]

            assert port.name == read_port.name
            assert port.trans == read_port.trans
            assert port.dcplx_trans == read_port.dcplx_trans
            assert port.port_type == read_port.port_type
            assert port.width == read_port.width
        assert wg_read.settings_units["length"] == "dbu"


def test_metainfo_read_cell(straight: kf.KCell) -> None:
    """Test whether we can read written metadata to a cell and its ports."""
    with NamedTemporaryFile("a", suffix=".oas") as t:
        save = kf.save_layout_options()
        save.write_context_info = True
        save.select_cell(straight.cell_index())
        straight.write(t.name, save_options=save)

        kcl = kf.KCLayout("TEST_META", infos=Layers)
        kcell = kcl.kcell(straight.name)
        kf.config.logfilter.regex = r"KLayout <=0.28.15 \(last update 2024-02-02\) cannot read LayoutMetaInfo on 'Cell.read'. kfactory uses these extensively for ports, info, and settings. Therefore proceed at your own risk."
        for cs in straight.kcl.cross_sections.cross_sections.values():
            kcl.get_cross_section(cs)
        kcell.read(t.name)
        kf.config.logfilter.regex = ""

        # TODO: wait for KLayout update https://github.com/KLayout/klayout/issues/1609

        # for i, port in enumerate(straight.ports):
        #     read_port = kcell.ports[i]

        #     assert port.name == read_port.name
        #     assert port.trans == read_port.trans
        #     assert port.dcplx_trans == read_port.dcplx_trans
        #     assert port.port_type == read_port.port_type
        #     assert port.width == read_port.width


def test_nometainfo_read(straight: kf.KCell) -> None:
    """Test whether we can turn of metadata writing."""
    with NamedTemporaryFile("a") as t:
        # save = kf.save_layout_options()
        # save.write_context_info = True
        straight.kcl.write(t.name, kf.save_layout_options(write_context_info=False))

        kcl = kf.KCLayout("TEST_META")
        kcl.read(t.name)

        wg_read = kcl[straight.name]
        wg_read.get_meta_data()

        assert wg_read.settings.model_dump() == {}
        assert len(wg_read.ports) == 0
        assert len(straight.ports) == 2
        assert straight.settings.model_dump() == {
            "length": 1000,
            "width": 500,
            "enclosure": "WGSTD",
            "layer": Layers().WG,
        }
        assert straight.function_name == "straight"
        assert straight.basename is None


def test_info_dump() -> None:
    c = kf.KCell()
    c.info = kf.Info(a="A")
    c.info.b = "B"  # type: ignore[attr-defined, unused-ignore]
    c.info["d"] = {"a": 1, "b": 2}
    c._settings = kf.KCellSettings(a="A", c="C")

    assert c.info == c.info.model_copy()
    assert c.settings == c.settings.model_copy()

    with NamedTemporaryFile("a") as t:
        # save = kf.save_layout_options()
        # save.write_context_info = True
        c.kcl.write(t.name)
        kcl = kf.KCLayout("TEST_META2")
        kcl.read(t.name)
        wg_read = kcl[c.name]
        wg_read.get_meta_data()
        assert wg_read.info == c.info
        assert wg_read.info["d"] == {"a": 1, "b": 2}



"""Straight virtual waveguide cells."""

from ...factories.virtual.straight import virtual_straight_factory
from ...kcell import kcl

virtual_straight = virtual_straight_factory(kcl)
"""Default straight on the "default" kcl."""



from functools import partial
import kfactory as kf
import numpy as np
from random import randint
from conftest import Layers


def test_all_angle_bundle(LAYER: Layers) -> None:
    sf = partial(kf.cells.virtual.straight.virtual_straight, layer=LAYER.WG)
    bf = partial(
        kf.cells.virtual.euler.virtual_bend_euler, layer=LAYER.WG, radius=10, width=1
    )

    # vc = kf.VKCell("test_all_angle")
    c = kf.KCell("test_all_angle")

    start_ports: list[kf.Port] = []
    end_ports: list[kf.Port] = []
    r = 50
    n = 3
    _l = 9

    for i in range(_l):
        # for i in range(1):
        a = (n - i) * 15
        a_rad = np.deg2rad(a)
        ae = 270 - n + i * 15
        ae_rad = np.deg2rad(ae)
        start_ports.append(
            c.create_port(
                name=f"s{i}",
                dcplx_trans=kf.kdb.DCplxTrans(
                    1, a, False, -500 + r * np.cos(a_rad), -100 + r * np.sin(a_rad)
                ),
                layer=c.kcl.find_layer(LAYER.WG),
                dwidth=1,
            )
        )
        end_ports.append(
            c.create_port(
                name=f"s{i + _l}",
                dcplx_trans=kf.kdb.DCplxTrans(
                    1, ae, False, 2510 + r * np.cos(ae_rad), 2410 + r * np.sin(ae_rad)
                ),
                layer=c.kcl.find_layer(LAYER.WG),
                dwidth=1,
            )
        )
    backbone = [
        kf.kdb.DPoint(750, -550),
        kf.kdb.DPoint(1000, 550),
        kf.kdb.DPoint(1000, 1200),
        kf.kdb.DPoint(2800, 1950),
    ]
    kf.routing.aa.optical.route_bundle(
        c,
        start_ports=start_ports,
        end_ports=end_ports,
        backbone=backbone,
        separation=[randint(1, 5) for _ in range(_l)],
        straight_factory=sf,
        bend_factory=bf,
    )



from layers import LAYER

import kfactory as kf


@kf.cell
def straight(width: int, length: int, width_exclude: int) -> kf.KCell:
    """Waveguide: Silicon on 1/0, Silicon exclude on 1/1"""
    c = kf.KCell()
    c.shapes(c.kcl.find_layer(LAYER.SI)).insert(kf.kdb.Box(0, -width // 2, length, width // 2))
    c.shapes(c.kcl.find_layer(LAYER.SIEXCLUDE)).insert(
        kf.kdb.Box(0, -width_exclude // 2, length, width_exclude // 2)
    )

    c.create_port(
        name="1", trans=kf.kdb.Trans(2, False, 0, 0), width=width, layer_info=LAYER.SI
    )
    c.create_port(
        name="2",
        trans=kf.kdb.Trans(0, False, length, 0),
        width=width,
        layer_info=LAYER.SI,
    )

    c.auto_rename_ports()

    return c


if __name__ == "__main__":
    kf.show(straight(2000, 50000, 5000))



import kfactory as kf
from conftest import Layers


@kf.cell
def simple_cplx_cell(layer: kf.kdb.LayerInfo) -> kf.KCell:
    c = kf.KCell()

    hh = 2.5

    dct = kf.kdb.DCplxTrans(1, 30, False, 2.5, 2.5)
    dbox = kf.kdb.DPolygon(kf.kdb.DBox(0, -hh, 10, hh)).transformed(dct)

    li = c.kcl.find_layer(layer)

    p1 = kf.Port(
        dwidth=hh * 2,
        layer=li,
        name="o1",
        dcplx_trans=dct * kf.kdb.DCplxTrans.R180,
    )
    p2 = kf.Port(
        dwidth=hh * 2,
        layer=li,
        name="o2",
        dcplx_trans=dct * kf.kdb.DCplxTrans(kf.kdb.DVector(10, 0)),
    )

    c.add_port(p1)
    c.add_port(p2)

    c.shapes(li).insert(dbox)

    c.draw_ports()
    return c


def test_cell(LAYER: Layers) -> None:
    simple_cplx_cell(LAYER.WG)


def test_connected_cell(LAYER: Layers) -> None:
    c = kf.KCell()
    layer = LAYER.WG
    sckc1 = c << simple_cplx_cell(layer)
    sckc2 = c << simple_cplx_cell(layer)
    sckc2.connect("o1", sckc1, "o1")



import kfactory as kf
import numpy as np
from layers import LAYER
import random


@kf.cell
def star(
    size: float, proportion: float, n_diamonds: int = 3, layer: kf.kdb.LayerInfo = LAYER.SI
) -> kf.KCell:
    """Create a diamond star cell

    Args:
        size: size in um
        proportion: width of the center vs size (0 < proportion <= 1)
        n_diamonds: number of diamonds in the star (>=1)

    Returns:
        Star Cell
    """

    c = kf.KCell()

    # the first star diamond, we use a box (int based)

    diamond = kf.kdb.DPolygon(
        [
            kf.kdb.DPoint(-size / 2, 0),
            kf.kdb.DPoint(0, -size / 2 * proportion),
            kf.kdb.DPoint(size / 2, 0),
            kf.kdb.DPoint(0, size / 2 * proportion),
        ]
    )

    li = c.kcl.find_layer(layer)
    c.shapes(li).insert(diamond)  # place base diamond

    for i in range(1, n_diamonds):
        angle = 180 / (n_diamonds) * i

        c.shapes(li).insert(
            diamond.transformed(kf.kdb.DCplxTrans(1, angle, False, 0, 0))
        )

    return c


@kf.cell
def merged_star(
    size: float, proportion: float, n_diamonds: int = 3, layer: kf.kdb.LayerInfo = LAYER.SI
) -> kf.KCell:
    """Same as star but use the star shapes and merge them to one polygon"""

    c = kf.KCell()

    _star = star(size, proportion, n_diamonds, layer)
    reg = kf.kdb.Region(_star.begin_shapes_rec(c.kcl.find_layer(layer)))
    reg.merge()  # merge the region

    # Insert the region

    c.shapes(c.kcl.find_layer(layer)).insert(reg)

    return c


@kf.cell
def sky_with_stars() -> kf.KCell:
    c = kf.KCell()

    box = kf.kdb.Box(0, 0, 400000, 400000)  # 400umx400um sky (default dbu)
    sky = kf.kdb.Region(box)

    # set a custom seed for random
    seed = 314159
    random.seed(seed)

    star_layer = LAYER.SI

    # create 50 stars
    for _ in range(50):
        x = random.uniform(10, 390)
        y = random.uniform(10, 390)
        angle = random.uniform(0, 360)
        _star = c << merged_star(
            size=random.uniform(5, 25),
            proportion=random.uniform(0.2, 0.3),
            n_diamonds=random.randint(2, 5),
            layer=star_layer,
        )
        _star.transform(kf.kdb.DTrans(angle, False, x, y))

    # remove the stars from the sky

    sky -= kf.kdb.Region(c.begin_shapes_rec(c.kcl.find_layer(star_layer)))

    c.shapes(c.kcl.find_layer(LAYER.SIEXCLUDE)).insert(sky)

    return c


if __name__ == "__main__":
    # kf.show(merged_star(5, 0.25))

    kf.show(sky_with_stars())



"""KFactory package. Utilities for creating photonic devices.

Uses the klayout package as a backend.

"""

# The import order matters, we need to first import the important stuff.
# isort:skip_file

__version__ = "0.21.1"

import klayout.db as kdb
import klayout.lay as lay
import klayout.rdb as rdb
from .kcell import (
    Info,
    Instance,
    InstanceGroup,
    KCLayout,
    KCell,
    KCellSettings,
    LayerEnum,
    LayerInfos,
    Constants,
    LayerStack,
    Port,
    Ports,
    VKCell,
    VInstance,
    cell,
    dpolygon_from_array,
    kcl,
    polygon_from_array,
    pprint_ports,
    save_layout_options,
    show,
)
from . import (
    cells,
    enclosure,
    kf_types,
    packing,
    placer,
    port,
    routing,
    technology,
    utils,
)
from .conf import config, logger
from .enclosure import LayerEnclosure, KCellEnclosure
from .grid import flexgrid_dbu, flexgrid, grid_dbu, grid

from aenum import constant  # type: ignore[import-untyped,unused-ignore]


__all__ = [
    "Constants",
    "Info",
    "Instance",
    "InstanceGroup",
    "KCLayout",
    "KCell",
    "KCellEnclosure",
    "KCellSettings",
    "LayerEnclosure",
    "LayerEnum",
    "LayerInfos",
    "LayerStack",
    "Port",
    "Ports",
    "VInstance",
    "VKCell",
    "cell",
    "cells",
    "config",
    "constant",
    "dpolygon_from_array",
    "enclosure",
    "flexgrid",
    "flexgrid_dbu",
    "grid",
    "grid_dbu",
    "kcl",
    "kdb",
    "lay",
    "logger",
    "packing",
    "placer",
    "polygon_from_array",
    "port",
    "pprint_ports",
    "rdb",
    "routing",
    "save_layout_options",
    "show",
    "technology",
    "kf_types",
    "utils",
]



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.2
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Enclosures

# %%
import kfactory as kf


class LayerInfos(kf.LayerInfos):
    WG: kf.kdb.LayerInfo = kf.kdb.LayerInfo(1, 0)
    SLAB: kf.kdb.LayerInfo = kf.kdb.LayerInfo(2, 0)
    NPP: kf.kdb.LayerInfo = kf.kdb.LayerInfo(3, 0)

LAYER = LayerInfos()
kf.kcl.infos = LAYER

# %%
enc = kf.enclosure.LayerEnclosure(
    [
        (LAYER.SLAB, 2000),
    ],
    name="WGSLAB",
    main_layer=LAYER.WG,
)
c = kf.cells.euler.bend_euler(radius=5, width=1, layer=LAYER.WG, enclosure=enc)
c.show()
c.plot()

# %%
enc = kf.enclosure.LayerEnclosure(
    [
        (LAYER.SLAB, 2000),
        (LAYER.NPP, 1000, 2000),
    ],
    name="SLAB_DOPED",
    main_layer=LAYER.WG,
)
c = kf.cells.euler.bend_euler(radius=5, width=1, layer=LAYER.WG, enclosure=enc)
c.show()
c.plot()

# %%
kcell_enc = kf.KCellEnclosure([enc])

@kf.cell
def two_bends(
    radius: float,
    width: float,
    layer: kf.kdb.LayerInfo,
    enclosure: kf.KCellEnclosure | None = None,
) -> kf.KCell:
    c = kf.KCell()
    b1 = c << kf.cells.euler.bend_euler(radius=radius, width=width, layer=layer)
    b2 = c << kf.cells.euler.bend_euler(radius=radius, width=width, layer=layer)
    b2.drotate(90)
    b2.dmovey(-11)
    b2.dmovex(b2.dxmin, 0)

    c.add_ports(b1.ports)
    c.add_ports(b2.ports)
    c.auto_rename_ports()

    if enclosure:
        enclosure.apply_minkowski_tiled(c)
    return c


c = two_bends(radius=5, width=1, layer=LAYER.WG, enclosure=kcell_enc)
c.show()
c.plot()

# %%



"""Provides straight waveguides in dbu and um versions.

A waveguide is a rectangle of material with excludes and/or slab around it::

    ┌─────────────────────────────┐
    │        Slab/Exclude         │
    ├─────────────────────────────┤
    │                             │
    │            Core             │
    │                             │
    ├─────────────────────────────┤
    │        Slab/Exclude         │
    └─────────────────────────────┘

The slabs and excludes can be given in the form of an
[Enclosure][kfactory.enclosure.LayerEncolosure].
"""

from .. import KCell, kcl, kdb, kf_types
from ..enclosure import LayerEnclosure
from ..factories.straight import straight_dbu_factory

__all__ = ["straight", "straight_dbu"]

straight_dbu = straight_dbu_factory(kcl)


def straight(
    width: kf_types.um,
    length: kf_types.um,
    layer: kdb.LayerInfo,
    enclosure: LayerEnclosure | None = None,
) -> KCell:
    """Straight waveguide in um.

    Visualization::

        ┌─────────────────────────────┐
        │        Slab/Exclude         │
        ├─────────────────────────────┤
        │                             │
        │            Core             │
        │                             │
        ├─────────────────────────────┤
        │        Slab/Exclude         │
        └─────────────────────────────┘

    Args:
        width: Width of the straight. [um]
        length: Length of the straight. [um]
        layer: Main layer of the straight.
        enclosure: Definition of slabs/excludes. [um]
    """
    return straight_dbu(
        width=round(width / kcl.dbu),
        length=round(length / kcl.dbu),
        layer=layer,
        enclosure=enclosure,
    )



"""Defines additional decorators than just `@cell`."""

from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Protocol, TypedDict, overload

from typing_extensions import Unpack

if TYPE_CHECKING:
    from cachetools import Cache

    from .kcell import (
        CHECK_INSTANCES,
        KC,
        KCell,
        KCellFunc,
        KCellParams,
        KCLayout,
        MetaData,
    )


class ModuleCellKWargs(TypedDict, total=False):
    """KWargs for `@module_cell`."""

    set_settings: bool
    set_name: bool
    check_ports: bool
    check_instances: CHECK_INSTANCES | None
    snap_ports: bool
    add_port_layers: bool
    cache: Cache[int, Any] | dict[int, Any] | None
    drop_params: list[str]
    register_factory: bool
    overwrite_existing: bool | None
    layout_cache: bool | None
    info: dict[str, MetaData] | None
    post_process: Iterable[Callable[[KCell], None]]
    debug_names: bool | None


class KCellDecoratorKWargs(TypedDict, total=False):
    """KWargs for `@cell`."""

    set_settings: bool
    set_name: bool
    check_ports: bool
    check_instances: CHECK_INSTANCES | None
    snap_ports: bool
    add_port_layers: bool
    cache: Cache[int, Any] | dict[int, Any] | None
    basename: str | None
    drop_params: list[str]
    register_factory: bool
    overwrite_existing: bool | None
    layout_cache: bool | None
    info: dict[str, MetaData] | None
    post_process: Iterable[Callable[[KC], None]]
    debug_names: bool | None


class KCellDecorator(Protocol):
    """Signature of the `@cell` decorator."""

    def __call__(
        self, **kwargs: Unpack[KCellDecoratorKWargs]
    ) -> Callable[[KCellFunc[KCellParams, KC]], KCellFunc[KCellParams, KC]]:
        """__call__ implementation."""


class ModuleDecorator(Protocol):
    """Signature of the `@module_cell` decorator."""

    def __call__(
        self, /, **kwargs: Unpack[ModuleCellKWargs]
    ) -> Callable[[KCellFunc[KCellParams, KC]], KCellFunc[KCellParams, KC]]:
        """__call__ implementation."""


def _module_cell(
    cell_decorator: KCellDecorator,
    /,
    **kwargs: Unpack[ModuleCellKWargs],
) -> Callable[[KCellFunc[KCellParams, KC]], KCellFunc[KCellParams, KC]]:
    """Constructs the actual decorator.

    Modifies the basename to the module if the module is not the main one.
    """

    def decorator_cell(
        f: KCellFunc[KCellParams, KC],
    ) -> KCellFunc[KCellParams, KC]:
        mod = f.__module__
        basename = f.__name__ if mod == "__main" else f"{mod}_{f.__name__}"
        return cell_decorator(basename=basename, **kwargs)(f)

    return decorator_cell


class Decorators:
    """Various decorators intended to be attached to a KCLayout."""

    def __init__(self, kcl: KCLayout):
        """Just set the standard `@cell` decorator."""
        self._cell = kcl.cell

    @overload
    def module_cell(
        self,
        _func: KCellFunc[KCellParams, KC],
        /,
    ) -> KCellFunc[KCellParams, KC]: ...

    @overload
    def module_cell(
        self, /, **kwargs: Unpack[ModuleCellKWargs]
    ) -> Callable[[KCellFunc[KCellParams, KC]], KCellFunc[KCellParams, KC]]: ...

    def module_cell(  # type: ignore[misc]
        self,
        _func: KCellFunc[KCellParams, KC] | None = None,
        /,
        **kwargs: Unpack[ModuleCellKWargs],
    ) -> (
        KCellFunc[KCellParams, KC]
        | Callable[[KCellFunc[KCellParams, KC]], KCellFunc[KCellParams, KC]]
    ):
        """Constructs the `@module_cell` decorator on KCLayout.decorators."""

        def mc(
            **kwargs: Unpack[ModuleCellKWargs],
        ) -> Callable[[KCellFunc[KCellParams, KC]], KCellFunc[KCellParams, KC]]:
            return _module_cell(self._cell, **kwargs)  # type: ignore[arg-type]

        return mc(**kwargs) if _func is None else mc(**kwargs)(_func)



import kfactory as kf
from collections.abc import Callable
from random import randint, uniform


def test_grid_dbu_1d(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            straight_factory_dbu(
                width=randint(500, 2500) * 2, length=randint(2000, 20_000)
            )
            for _ in range(10)
        ],
        spacing=5000,
        align_x="origin",
    )


def test_grid_dbu_2d(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        spacing=5000,
        align_x="origin",
    )


def test_grid_dbu_2d_uneven(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for i in range(10 + j**2)
            ]
            for j in range(-3, 3)
        ],
        spacing=5000,
        align_x="xmin",
        align_y="ymax",
    )


def test_grid_dbu_2d_rotation(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        rotation=1,
        spacing=5000,
        align_x="xmin",
        align_y="ymin",
    )


def test_grid_dbu_1d_shape(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            straight_factory_dbu(
                width=randint(500, 2500) * 2, length=randint(2000, 20_000)
            )
            for _ in range(10)
        ],
        spacing=5000,
        align_x="origin",
        shape=(1, 10),
    )


def test_grid_dbu_2d_shape(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        spacing=5000,
        align_x="origin",
        shape=(4, 5),
    )


def test_grid_dbu_2d_shape_rotation(
    straight_factory_dbu: Callable[..., kf.KCell],
) -> None:
    c = kf.KCell()

    kf.grid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        rotation=1,
        spacing=5000,
        align_x="origin",
        shape=(3, 7),
    )


def test_flexgrid_dbu_1d(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.flexgrid_dbu(
        c,
        kcells=[
            straight_factory_dbu(
                width=randint(500, 2500) * 2, length=randint(2000, 20_000)
            )
            for _ in range(10)
        ],
        spacing=5000,
        align_x="origin",
    )


def test_flexgrid_dbu_2d(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.flexgrid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        spacing=5000,
        align_x="origin",
    )


def test_flexgrid_dbu_2d_rotation(
    straight_factory_dbu: Callable[..., kf.KCell],
) -> None:
    c = kf.KCell()

    kf.flexgrid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        rotation=1,
        spacing=5000,
        align_x="center",
    )


def test_flexgrid_dbu_1d_shape(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.flexgrid_dbu(
        c,
        kcells=[
            straight_factory_dbu(
                width=randint(500, 2500) * 2, length=randint(2000, 20_000)
            )
            for _ in range(10)
        ],
        spacing=5000,
        align_x="origin",
        shape=(1, 10),
    )


def test_flexgrid_dbu_2d_shape(straight_factory_dbu: Callable[..., kf.KCell]) -> None:
    c = kf.KCell()

    kf.flexgrid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        spacing=5000,
        align_x="origin",
        shape=(4, 5),
    )


def test_flexgrid_dbu_2d_shape_rotation(
    straight_factory_dbu: Callable[..., kf.KCell],
) -> None:
    c = kf.KCell()

    kf.flexgrid_dbu(
        c,
        kcells=[
            [
                straight_factory_dbu(
                    width=randint(500, 2500) * 2, length=randint(2000, 20_000)
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        rotation=1,
        spacing=5000,
        align_x="origin",
        shape=(3, 7),
    )


def test_grid_2d_shape_rotation(
    straight_factory: Callable[..., kf.KCell],
) -> None:
    c = kf.KCell()

    kf.grid(
        c,
        kcells=[
            [
                straight_factory(
                    width=round(uniform(0.5, 2.5)) * 2, length=round(uniform(2, 20))
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        rotation=1,
        spacing=5,
        align_x="origin",
        shape=(3, 7),
        target_trans=kf.kdb.DCplxTrans(1, 37, False, 0, 0),
    )


def test_flexgrid_2d_shape_rotation(
    straight_factory: Callable[..., kf.KCell],
) -> None:
    c = kf.KCell()

    kf.flexgrid(
        c,
        kcells=[
            [
                straight_factory(
                    width=round(uniform(0.5, 2.5)) * 2, length=round(uniform(2, 20))
                )
                for _ in range(10)
            ]
            for _ in range(2)
        ],
        rotation=1,
        spacing=5,
        align_x="origin",
        shape=(3, 7),
        target_trans=kf.kdb.DCplxTrans(1, 37, False, 0, 0),
    )



"""Generic routing functions which are independent of the potential use."""

from __future__ import annotations

from collections import defaultdict
from collections.abc import Iterable, Sequence
from typing import Any, Literal, Protocol

from pydantic import BaseModel, Field

from .. import kdb, rdb
from ..conf import config
from ..kcell import Instance, KCell, Port
from ..kf_types import dbu
from .manhattan import ManhattanBundleRoutingFunction, ManhattanRouter, route_smart

__all__ = [
    "PlacerFunction",
    "ManhattanRoute",
    "check_collisions",
    "get_radius",
    "route_bundle",
]


class PlacerFunction(Protocol):
    """A placer function. Used to place Instances given a path."""

    def __call__(
        self,
        c: KCell,
        p1: Port,
        p2: Port,
        pts: Sequence[kdb.Point],
        route_width: dbu | None = None,
        **kwargs: Any,
    ) -> ManhattanRoute:
        """Implementation of the function."""


class RouterPostProcessFunction(Protocol):
    """A function that can be used to post process functions."""

    def __call__(
        self,
        *,
        c: KCell,
        routers: list[ManhattanRouter],
        start_ports: list[Port],
        end_ports: list[Port],
        **kwargs: Any,
    ) -> None:
        """Implementation of post process function."""


class ManhattanRoute(BaseModel, arbitrary_types_allowed=True):
    """Optical route containing a connection between two ports.

    Attrs:
        backbone: backbone points
        start_port: port at the first instance denoting the start of the route
        end_port: port at the last instance denoting the end of the route
        instances: list of the instances in order from start to end of the route
        n_bend90: number of bends used
        length: length of the route without the bends
        length_straights: length of the straight_factory elements
    """

    backbone: list[kdb.Point]
    start_port: Port
    end_port: Port
    instances: list[Instance] = Field(default_factory=list)
    n_bend90: int = 0
    n_taper: int = 0
    bend90_radius: dbu = 0
    taper_length: dbu = 0
    length: dbu = 0
    """Length of backbone without the bends."""
    length_straights: dbu = 0
    polygons: dict[kdb.LayerInfo, list[kdb.Polygon]] = Field(default_factory=dict)

    @property
    def length_backbone(self) -> dbu:
        """Length of the backbone in dbu."""
        length = 0
        p_old = self.backbone[0]
        for p in self.backbone[1:]:
            length += int((p - p_old).length())
            p_old = p
        return length


def check_collisions(
    c: KCell,
    start_ports: list[Port],
    end_ports: list[Port],
    routers: list[ManhattanRouter],
    routes: list[ManhattanRoute],
    on_collision: Literal["error", "show_error"] | None = "show_error",
    collision_check_layers: Sequence[kdb.LayerInfo] | None = None,
) -> None:
    """Checks for collisions given manhattan routes.

    Args:
        c: The KCell to check.
        start_ports: Ports from which the routes are supposed to start.
        end_ports: Ports where the routes are supposed to end.
        routers: The ManhattanRouters that constructed the routes.
        routes: The ManhatnnaRoutes which were used by the placer.
        on_collision: What to do on error. Can either do nothing (None),
            throw an error ("error"), or throw an error and open the
            cell with report in Klayout ("show_error").
        collision_check_layers: Sequence of layers which should be checked for
            overlaps to determine error. If not defined, all layers occurring in
            ports will be used.
    """
    if on_collision is not None:
        collision_edges: dict[str, kdb.Edges] = {}
        inter_route_collisions = kdb.Edges()
        all_router_edges = kdb.Edges()
        for i, (ps, pe, router) in enumerate(zip(start_ports, end_ports, routers)):
            _edges, router_edges = router.collisions(log_errors=None)
            if not _edges.is_empty():
                collision_edges[f"{ps.name} - {pe.name} (index: {i})"] = _edges
            inter_route_collision = all_router_edges.interacting(router_edges)
            if not inter_route_collision.is_empty():
                inter_route_collisions.join_with(inter_route_collision)
            all_router_edges.join_with(router_edges)

        if collision_edges or not inter_route_collisions.is_empty():
            if collision_check_layers is None:
                collision_check_layers = list(set(p.layer_info for p in start_ports))
            dbu = c.kcl.dbu
            db = rdb.ReportDatabase("Routing Errors")
            cat = db.create_category("Manhattan Routing Collisions")
            c.name = c.name[: config.max_cellname_length]
            cell = db.create_cell(c.name)
            for name, edges in collision_edges.items():
                item = db.create_item(cell, cat)
                item.add_value(name)
                for edge in edges.each():
                    item.add_value(edge.to_dtype(dbu))
            insts = [inst for route in routes for inst in route.instances]
            shapes: dict[kdb.LayerInfo, list[kdb.Region]] = defaultdict(list)
            for route in routes:
                for layer, _shapes in route.polygons.items():
                    shapes[layer].append(kdb.Region(_shapes))
            layer_cats: dict[kdb.LayerInfo, rdb.RdbCategory] = {}

            def layer_cat(layer_info: kdb.LayerInfo) -> rdb.RdbCategory:
                if layer_info not in layer_cats:
                    layer_cats[layer_info] = db.category_by_path(
                        layer_info.to_s()
                    ) or db.create_category(layer_info.to_s())
                return layer_cats[layer_info]

            any_layer_collision = False

            for layer_info in collision_check_layers:
                shapes_regions = shapes[layer_info]
                layer = c.kcl.layer(layer_info)
                error_region_instances = kdb.Region()
                error_region_shapes = kdb.Region()
                inst_regions: dict[int, kdb.Region] = {}
                inst_region = kdb.Region()
                shape_region = kdb.Region()
                for r in shapes_regions:
                    if not (shape_region & r).is_empty():
                        error_region_shapes.insert(shape_region & r)
                    shape_region.insert(r)
                for i, inst in enumerate(insts):
                    _inst_region = kdb.Region(inst.bbox(layer))
                    # inst_shapes: kdb.Region | None = None
                    if not (inst_region & _inst_region).is_empty():
                        # if inst_shapes is None:
                        inst_shapes = kdb.Region()
                        shape_it = c.begin_shapes_rec_overlapping(
                            layer, inst.bbox(layer)
                        )
                        shape_it.select_cells([inst.cell.cell_index()])
                        shape_it.min_depth = 1
                        for _it in shape_it.each():
                            if _it.path()[0].inst() == inst._instance:
                                inst_shapes.insert(
                                    _it.shape().polygon.transformed(_it.trans())
                                )
                        for j, _reg in inst_regions.items():
                            if _reg & _inst_region:
                                __reg = kdb.Region()
                                shape_it = c.begin_shapes_rec_touching(
                                    layer, (_reg & _inst_region).bbox()
                                )
                                shape_it.select_cells([insts[j].cell.cell_index()])
                                shape_it.min_depth = 1
                                for _it in shape_it.each():
                                    if _it.path()[0].inst() == insts[j]._instance:
                                        __reg.insert(
                                            _it.shape().polygon.transformed(_it.trans())
                                        )

                                error_region_instances.insert(__reg & inst_shapes)
                    inst_region += _inst_region
                    inst_regions[i] = _inst_region

                if not error_region_shapes.is_empty():
                    any_layer_collision = True
                    if on_collision == "error":
                        continue
                    cat = layer_cat(layer_info)
                    sc = db.category_by_path(
                        f"{cat.path()}.RoutingErrors"
                    ) or db.create_category(layer_cat(layer_info), "RoutingErrors")
                    for poly in error_region_shapes.merge().each():
                        it = db.create_item(cell, sc)
                        it.add_value("Route shapes overlapping with other shapes")
                        it.add_value(c.kcl.to_um(poly))
                if not error_region_instances.is_empty():
                    any_layer_collision = True
                    if on_collision == "error":
                        continue
                    cat = layer_cat(layer_info)
                    sc = db.category_by_path(
                        f"{cat.path()}.RoutingErrors"
                    ) or db.create_category(layer_cat(layer_info), "RoutingErrors")
                    for poly in error_region_instances.merge().each():
                        it = db.create_item(cell, sc)
                        it.add_value("Route instances overlapping with other instances")
                        it.add_value(c.kcl.to_um(poly))

            if any_layer_collision:
                match on_collision:
                    case "show_error":
                        c.show(lyrdb=db)
                        raise RuntimeError(
                            f"Routing collision in {c.kcl.future_cell_name or c.name}"
                        )
                    case "error":
                        raise RuntimeError(
                            f"Routing collision in {c.kcl.future_cell_name or c.name}"
                        )


def get_radius(
    ports: Iterable[Port],
) -> dbu:
    """Calculates a radius between two ports.

    This can be used to determine the radius of two bend ports.
    """
    ports = tuple(ports)
    if len(ports) != 2:
        raise ValueError(
            "Cannot determine the maximal radius of a bend with more than two ports."
        )
    p1, p2 = ports
    if p1.angle == p2.angle:
        return int((p1.trans.disp - p2.trans.disp).length())
    _p = kdb.Point(1, 0)
    e1 = kdb.Edge(p1.trans.disp.to_p(), p1.trans * _p)
    e2 = kdb.Edge(p2.trans.disp.to_p(), p2.trans * _p)

    center = e1.cut_point(e2)
    if center is None:
        raise ValueError("Could not determine the radius. Something went very wrong.")
    return int(
        max((p1.trans.disp - center).length(), (p2.trans.disp - center).length())
    )


def route_bundle(
    *,
    c: KCell,
    start_ports: list[Port],
    end_ports: list[Port],
    start_straights: dbu | list[dbu] = 0,
    end_straights: dbu | list[dbu] = 0,
    route_width: dbu | list[dbu] | None = None,
    sort_ports: bool = False,
    on_collision: Literal["error", "show_error"] | None = "show_error",
    collision_check_layers: Sequence[kdb.LayerInfo] | None = None,
    routing_function: ManhattanBundleRoutingFunction = route_smart,
    routing_kwargs: dict[str, Any] = {"bbox_routing": "minimal"},
    placer_function: PlacerFunction,
    placer_kwargs: dict[str, Any] = {},
    router_post_process_function: RouterPostProcessFunction | None = None,
    router_post_process_kwargs: dict[str, Any] = {},
) -> list[ManhattanRoute]:
    """Route a bundle from starting ports to end_ports.

    Args:
        c: Cell to place the route in.
        start_ports: List of start ports.
        end_ports: List of end ports.
        separation: Separation between the routes.
        start_straights: Minimal straight segment after `p1`.
        end_straights: Minimal straight segment before `p2`.
        collision_check_layers: Layers to check for actual errors if manhattan routes
            detect potential collisions.
        on_collision: Define what to do on routing collision. Default behaviour is to
            open send the layout of c to klive and open an error lyrdb with the
            collisions. "error" will simply raise an error. None will ignore any error.
        bboxes: List of boxes to consider. Currently only boxes overlapping ports will
            be considered.
        route_width: Width of the route. If None, the width of the ports is used.
        sort_ports: Automatically sort ports.
        bbox_routing: "minimal": only route to the bbox so that it can be safely routed
            around, but start or end bends might encroach on the bounding boxes when
            leaving them.
        bend90_radius: The radius with which the router will try to router. This should
            normally be the maximal radius used.
        placer_function: Function to place the routes. Must return a corresponding list
            of OpticalManhattan routes.
            Must accept the following protocol:
            ```
            placer_function(
                c: KCell, p1: Port, p2: Port, pts: list[Point], **placer_kwargs
            )
            ```
        placer_kwargs: Additional kwargs passed to the placer_function.
        routing_function: Function to place the routes. Must return a corresponding list
            of OpticalManhattan routes.
            Must accept the following protocol:
            ```
            routing_function(
                c: KCell, p1: Port, p2: Port, pts: list[Point], **placer_kwargs
            )
            ```
        routing_kwargs: Additional kwargs passed to the placer_function.
        router_post_process_function: Function used to modify the routers returned by
            the routing function. This is particularly useful for operations such as
            path length matching.
        router_post_process_kwargs: Kwargs for router_post_process_function.
    """
    if not start_ports:
        return []
    if not (len(start_ports) == len(end_ports)):
        raise ValueError(
            "For bundle routing the input port list must have"
            " the same size as the end ports and be the same length."
        )

    if isinstance(start_straights, int):
        start_straights = [start_straights] * len(start_ports)
    if isinstance(end_straights, int):
        end_straights = [end_straights] * len(start_ports)

    if route_width:
        if isinstance(route_width, int):
            widths = [route_width] * len(start_ports)
        else:
            widths = route_width
    else:
        widths = [p.width for p in start_ports]

    routers = routing_function(
        start_ports=start_ports,
        end_ports=end_ports,
        widths=widths,
        start_straights=start_straights,
        end_straights=end_straights,
        **routing_kwargs,
    )

    if not routers:
        return []

    start_mapping = {sp.trans: sp for sp in start_ports}
    end_mapping = {ep.trans: ep for ep in end_ports}
    routes: list[ManhattanRoute] = []
    start_ports = []
    end_ports = []

    for router in routers:
        sp = start_mapping[router.start_transformation]
        ep = end_mapping[router.end_transformation]
        start_ports.append(sp)
        end_ports.append(ep)

    if router_post_process_function is not None:
        router_post_process_function(
            c=c,
            start_ports=start_ports,
            end_ports=end_ports,
            routers=routers,
            **router_post_process_kwargs,
        )
    for router, ps, pe in zip(routers, start_ports, end_ports):
        routes.append(
            placer_function(
                c,
                ps,
                pe,
                router.start.pts,
                **placer_kwargs,
            )
        )
    check_collisions(
        c=c,
        start_ports=start_ports,
        end_ports=end_ports,
        on_collision=on_collision,
        collision_check_layers=collision_check_layers,
        routers=routers,
        routes=routes,
    )
    return routes



"""Kfactory configuration."""

from __future__ import annotations

import importlib
import os
import re
import sys
import traceback
from enum import Enum
from itertools import takewhile
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Protocol, cast, runtime_checkable

import loguru
import rich.console
from dotenv import find_dotenv
from loguru import logger as logger
from pydantic import BaseModel, Field, ValidationError, ValidationInfo, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from . import kdb, rdb

if TYPE_CHECKING:
    from .kcell import KCell, KCLayout


@runtime_checkable
class ShowFunction(Protocol):
    def __call__(
        self,
        layout: KCLayout | KCell | Path | str,
        *,
        lyrdb: rdb.ReportDatabase | Path | str | None,
        l2n: kdb.LayoutToNetlist | Path | str | None,
        keep_position: bool,
        save_options: kdb.SaveLayoutOptions,
        use_libraries: bool,
        library_save_options: kdb.SaveLayoutOptions,
    ) -> None: ...


def add_traceback(record: loguru.Record) -> None:
    """Add a traceback to the logger."""
    extra = record["extra"]
    if extra.get("with_traceback", False):
        extra["traceback"] = "\n" + "".join(traceback.format_stack())
    else:
        extra["traceback"] = ""


def tracing_formatter(record: loguru.Record) -> str:
    """Traceback filtering.

    Filter out frames coming from Loguru internals.
    """
    frames = takewhile(
        lambda f: "/loguru/" not in f.filename, traceback.extract_stack()
    )
    stack = " > ".join(f"{f.filename}:{f.name}:{f.lineno}" for f in frames)
    record["extra"]["stack"] = stack

    if record["extra"].get("with_backtrace", False):
        return (
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level>"
            " | <cyan>{extra[stack]}</cyan> - <level>{message}</level>\n{exception}"
        )

    else:
        return (
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}"
            "</level> | <cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan>"
            " - <level>{message}</level>\n{exception}"
        )


class LogLevel(str, Enum):
    """KFactory logger levels."""

    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class CHECK_INSTANCES(str, Enum):
    RAISE = "error"
    FLATTEN = "flatten"
    VINSTANCES = "vinstances"
    IGNORE = "ignore"


class LogFilter(BaseModel):
    """Filter certain messages by log level or regex.

    Filtered messages are not evaluated and discarded.
    """

    level: LogLevel = LogLevel.INFO
    regex: str | None = None

    def __call__(self, record: loguru.Record) -> bool:
        """Loguru needs the filter to be callable."""
        levelno = logger.level(self.level).no
        if self.regex is None:
            return record["level"].no >= levelno
        else:
            return record["level"].no >= levelno and not bool(
                re.search(self.regex, record["message"])
            )


def get_show_function(value: str | ShowFunction) -> ShowFunction:
    if isinstance(value, str):
        mod, f = value.rsplit(".", 1)
        loaded_mod = importlib.import_module(mod)
        return loaded_mod.__getattribute__(f)  # type: ignore[no-any-return]
    return value


def get_affinity() -> int:
    """Get number of cores/threads available.

    On (most) linux we can get it through the scheduling affinity. Otherwise,
    fall back to the multiprocessing cpu count.
    """
    try:
        threads = len(os.sched_getaffinity(0))  # type: ignore[attr-defined,unused-ignore]
    except AttributeError:
        import multiprocessing

        threads = multiprocessing.cpu_count()
    finally:
        return threads


dotenv_path = find_dotenv(usecwd=True)


class Settings(BaseSettings):
    """KFactory settings object.

    Attrs:
        n_threads: Number of threads to use for tiling processor, defaults
            to all available cores.
        logger: The loguru class to use for logging. Shouldn't be necessary
            to configure by hand.
        logfilter: The filter to use. Can be configured to set log level and filter
            messages by regex.
        display_type: The type of image to show when calling the jupyter display
            function.
        meta_format: Metadata format to use for reading KLayout metadata.
            If set to 'default', metadata will be read as instances such as
            Trans/DCplxTrans. If the metadata is in the old string format
            (there was a bug in how to read metadata in some versions), use
            'string'.
        console: The rich console to use for displaying rich content.
        max_cellname_length: The maximum length of a cell name.
    """

    model_config = SettingsConfigDict(
        arbitrary_types_allowed=True,
        env_prefix="kfactory_",
        env_nested_delimiter="_",
        extra="allow",
        validate_assignment=True,
        env_file=dotenv_path,
    )

    n_threads: int = get_affinity()
    """Number of threads to use for multithreaded operations."""
    logfilter: LogFilter = Field(default_factory=LogFilter)
    """Can configure the logger to ignore certain levels or by regex."""
    display_type: Literal["widget", "image"] = "image"
    """The default behavior for displaying cells in jupyter."""
    meta_format: Literal["v3", "v2", "v1"] = "v3"
    """The format of the saving of metadata.

    v1: Transformations and other KLayout objects are stored as a string. In
        case of ports they are converted back to KLayout objects on read.
    v2: All objects can be stored in the nativ KLayout format (klayout>=0.28.13)
    """
    # console for printing
    console: rich.console.Console = Field(default_factory=rich.console.Console)

    # cell decorator settings
    allow_width_mismatch: bool = False
    allow_layer_mismatch: bool = False
    allow_type_mismatch: bool = False
    allow_undefined_layers: bool = False
    cell_layout_cache: bool = False
    cell_overwrite_existing: bool = False
    connect_use_angle: bool = True
    connect_use_mirror: bool = True
    check_instances: CHECK_INSTANCES = CHECK_INSTANCES.RAISE
    max_cellname_length: int = 99
    debug_names: bool = False

    # default write settings
    write_cell_properties: bool = True
    write_context_info: bool = True
    write_file_properties: bool = True

    show_function: ShowFunction | None = None

    @field_validator("show_function", mode="before")
    @classmethod
    def _validate_show_function(
        cls, show: ShowFunction | str | None
    ) -> ShowFunction | None:
        if isinstance(show, str):
            mod, f = show.rsplit(".", 1)
            loaded_mod = importlib.import_module(mod)
            _show = cast(ShowFunction, loaded_mod.__getattribute__(f))
            if not isinstance(_show, ShowFunction):
                raise ValidationError(f"{show=} is not a ShowFunction.")
            show = _show
        return show

    @field_validator("logfilter")
    @classmethod
    def _validate_logfilter(cls, logfilter: LogFilter) -> LogFilter:
        logger.remove()
        logger.add(
            sys.stdout,
            format=tracing_formatter,
            filter=logfilter,
            enqueue=True,
            backtrace=True,
        )
        logger.debug("LogLevel: {}", logfilter.level)
        logger.patch(add_traceback)

        return logfilter

    @field_validator("cell_overwrite_existing")
    @classmethod
    def _validate_overwrite_and_cache(cls, v: bool, info: ValidationInfo) -> bool:
        if v is True:
            logger.warning(
                "'overwrite_existing' has been set to True. This might cause "
                "unintended behavior when overwriting existing cells and delete any "
                "existing instances of them."
            )
        return v

    @field_validator("cell_layout_cache")
    @classmethod
    def _validate_layout_cache(cls, v: bool, info: ValidationInfo) -> bool:
        if v is True:
            logger.debug(
                "'cell_layout_cache' has been set to True. This might cause when "
                "as any cell names generated automatically are loaded from the layout"
                " instead of created. This could happen e.g. after reading a gds file"
                " into the layout."
            )
        return v

    @field_validator(
        "allow_width_mismatch", "allow_layer_mismatch", "allow_type_mismatch"
    )
    @classmethod
    def _debug_info_on_global_setting(cls, v: Any, info: ValidationInfo) -> Any:
        logger.bind(with_traceback=True).debug(
            "'{}' set globally to '{}'", info.field_name, v
        )
        return v

    def __init__(self, **data: Any):
        """Set log filter and run pydantic."""
        super().__init__(**data)


config = Settings()


__all__ = ["config", "LogLevel"]
